import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack.mathfun.deriv_tst import test_num_deriv
from TRXASprefitpack import edge_gaussian, edge_lorenzian
from TRXASprefitpack import deriv_edge_gaussian, deriv_edge_lorenzian
from TRXASprefitpack import voigt, voigt_thy
from TRXASprefitpack import deriv_voigt, deriv_voigt_thy

class TestDerivPeakShape(unittest.TestCase):

    def test_deriv_voigt_1(self):
        '''
        Test gradient of voigt function (when fwhm_G == 0)
        '''
        e = np.linspace(-5, 5, 1001)
        fwhm_L = 1
        tst = deriv_voigt(e, 0, fwhm_L)
        ref = test_num_deriv(voigt, e, 0, fwhm_L)
        result = np.allclose(tst[:, [0, 2]], ref[:, [0, 2]])
        self.assertEqual(result, True)

    def test_deriv_voigt_2(self):
        '''
        Test gradient of voigt function (when fwhm_L == 0)
        '''
        e = np.linspace(-5, 5, 1001)
        fwhm_G = 1
        tst = deriv_voigt(e, fwhm_G, 0)
        ref = test_num_deriv(voigt, e, fwhm_G, 0)
        result = np.allclose(tst[:, [0, 1]], ref[:, [0, 1]])
        self.assertEqual(result, True)

    def test_deriv_voigt_3(self):
        '''
        Test gradient of voigt function (when fwhm_L == 0)
        '''
        e = np.linspace(-5, 5, 1001)
        fwhm_G = 1; fwhm_L = 0.5
        tst = deriv_voigt(e, fwhm_G, fwhm_L)
        ref = test_num_deriv(voigt, e, fwhm_G, fwhm_L)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)
    
    def test_deriv_edge_1(self):
        '''
        Test gradient of gaussian type edge
        '''
        e = np.linspace(-5, 5, 1001)
        fwhm_G = 1
        tst = deriv_edge_gaussian(e, fwhm_G)
        ref = test_num_deriv(edge_gaussian,
        e, fwhm_G)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)

    def test_deriv_edge_2(self):
        '''
        Test gradient of lorenzian type edge
        '''
        e = np.linspace(-5, 5, 1001)
        fwhm_L = 1
        tst = deriv_edge_lorenzian(e, fwhm_L)
        ref = test_num_deriv(edge_lorenzian,
        e, fwhm_L)
        result = np.allclose(tst, ref)
        self.assertEqual(result, True)
    
    def test_deriv_voigt_thy_1(self):
        '''
        Test gradient of voigt_thy (policy == 'shift')
        '''
        e = np.linspace(-10, 20, 300)
        thy_peak = \
            np.array([[-3, -2, 1, 0, 3, 3.1, 3.5, 4, 10, 15], 
            [1e-2, 1e-2, 1e-1, 2e-2, 5e-2, 3e-2, 6e-2, 1e-3, 1e-2, 1.5e-2]]).T
        tst = deriv_voigt_thy(e, thy_peak, 0.3, 0.5, -1.5, 'shift')
        ref = test_num_deriv(lambda e, fwhm_G, fwhm_L, peak_shift: \
            voigt_thy(e, thy_peak, fwhm_G, fwhm_L, peak_shift, 'shift'),
            e, 0.3, 0.5, -1.5)
        result = np.allclose(tst, ref[:, 1:])
        self.assertEqual(result, True)

    def test_deriv_voigt_thy_2(self):
        '''
        Test gradient of voigt_thy (policy == 'scale')
        '''
        e = np.linspace(-10, 20, 300)
        thy_peak = \
            np.array([[-3, -2, 1, 0, 3, 3.1, 3.5, 4, 10, 15], 
            [1e-2, 1e-2, 1e-1, 2e-2, 5e-2, 3e-2, 6e-2, 1e-3, 1e-2, 1.5e-2]]).T
        tst = deriv_voigt_thy(e, thy_peak, 0.3, 0.5, 0.996, 'scale')
        ref = test_num_deriv(lambda e, fwhm_G, fwhm_L, peak_scale: \
            voigt_thy(e, thy_peak, fwhm_G, fwhm_L, peak_scale, 'scale'),
            e, 0.3, 0.5, 0.996)
        result = np.allclose(tst, ref[:, 1:])
        self.assertEqual(result, True)

    def test_deriv_voigt_thy_3(self):
        '''
        Test gradient of voigt_thy (policy == 'both')
        '''
        e = np.linspace(-10, 20, 300)
        thy_peak = \
            np.array([[-3, -2, 1, 0, 3, 3.1, 3.5, 4, 10, 15], 
            [1e-2, 1e-2, 1e-1, 2e-2, 5e-2, 3e-2, 6e-2, 1e-3, 1e-2, 1.5e-2]]).T
        tst = deriv_voigt_thy(e, thy_peak, 0.3, 0.5, 
        np.array([-1.5, 0.996]), 'both')
        ref = test_num_deriv(lambda e, fwhm_G, fwhm_L, peak_shift, peak_scale: \
            voigt_thy(e, thy_peak, fwhm_G, fwhm_L, 
            np.array([peak_shift, peak_scale]), 'both'),
            e, 0.3, 0.5, -1.5, 0.996)
        result = np.allclose(tst, ref[:, 1:])
        self.assertEqual(result, True)




if __name__ == '__main__':
    unittest.main()



# misc
# submodule for miscellaneous function of
# tools subpackage

from typing import Tuple
import numpy as np

def read_data(prefix: str, num_scan: int, num_data_pts: int, default_SN: float) -> Tuple[np.ndarray, np.ndarray]:
    '''
    Read data from prefix_i.txt (1 <= i <= num_scan)

    Args:
     prefix: prefix of scan_file
     num_scan: the number of scan file to read
     num_data_pts: the number of data points per each scan file
     default_SN: Default Signal/Noise
    
    Return:
     Tuple of data of eps
    '''
    data = np.zeros((num_data_pts, num_scan))
    eps = np.zeros((num_data_pts, num_scan))

    for i in range(num_scan):
        A = np.genfromtxt(f'{prefix}_{i+1}.txt')
        data[:, i] = A[:, 1]
        if A.shape[1] == 2:
            eps[:, i] = np.max(np.abs(data[:, i]))*np.ones(num_data_pts)/default_SN
        else:
            eps[:, i] = A[:, 2]
    return data, eps

def parse_matrix(mat_str: np.ndarray, tau_rate: np.ndarray) -> np.ndarray:
    '''
    Parse user supplied rate equation matrix

    Args:
     mat_str: user supplied rate equation (lower triangular matrix)
     tau_rate: time constants for rate equation parameter (1/ki)
    
    Return:
     parsed rate equation matrix.
     the value of lifetime 
     parameters (1/ki) used to define rate equation matrix
    
    Note:
     Every entry in the rate equation matrix should be
     '0', '1*ki', '-x.xxx*ki', 'x.xxx*ki' or '-(x.xxx*ki+y.yyy*kj+...)'
     Number of non zero diagonal elements and size of tau should be same
     Number of parameter used to define rate equation and size of tau should 
     be same.
    '''

    L = np.zeros_like(mat_str, dtype=float)
    mask = (mat_str != '0')
    red_mat_str = mat_str[mask]
    red_L = np.zeros_like(red_mat_str, dtype=float)

    for i in range(red_mat_str.size):
        tmp = red_mat_str[i]
        if '-' in tmp:
            tmp = tmp.strip('-')
            tmp = tmp.strip('('); tmp = tmp.strip(')')
            k_lst = tmp.split('+')
            for k in k_lst:
                k_pair = k.split('*')
                red_L[i] = red_L[i] - float(k_pair[0])/tau_rate[int(k_pair[1][1:])-1]
        else:
            tmp_pair = tmp.split('*')
            red_L[i] = float(tmp_pair[0])/tau_rate[int(tmp_pair[1][1:])-1]
    
    L[mask] = red_L
    
    return L


'''
res:
subpackage for resdiual function for
fitting time delay scan data or static spectrum data

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from .parm_bound import set_bound_t0, set_bound_tau
from .parm_bound import set_bound_e0
from .res_decay import residual_decay, res_grad_decay
from .res_osc import residual_dmp_osc, res_grad_dmp_osc
from .res_both import residual_both, res_grad_both
from .res_voigt import residual_voigt, res_grad_voigt
from .res_thy import residual_thy, res_grad_thy

__all__ = ['set_bound_t0', 'set_bound_tau',
           'set_bound_e0',
           'residual_decay', 'res_grad_decay',
           'residual_dmp_osc', 'res_grad_dmp_osc',
           'residual_both', 'res_grad_both',
           'residual_voigt', 'res_grad_voigt',
           'residual_thy', 'res_grad_thy']


'''
static_result:
submodule for reporting fitting process of static spectrum result

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Callable
import numpy as np
from scipy.special import wofz
import h5py as h5
from ..mathfun.peak_shape import voigt, \
      edge_gaussian, edge_lorenzian

class StaticResult(dict):
      '''
      Represent results for fitting static driver routine

      Attributes:
       model ({'thy', 'voigt'}): model used for fitting

              `voigt`: sum of voigt function, edge function and base funtion

              `thy`: sum of voigt broadened theoretical lineshape spectrum, edge function and base function
       thy_peak (sequence of np.ndarray): theoretical calculated peak position and intensity [model: `thy`]
       policy ({'shift', 'scale', 'both'}): policy to match discrepancy between theoretical spectrum and
        experimental static spectrum.
       e (np.ndarray): energy range
       intensity (np.ndarray): intensity of static spectrum
       eps (np.ndarray): estimated error of static spectrum
       fit (np.ndarray): fitting curve for data (n,)
       fit_comp (np.ndarray): curve for each voigt component and edge
       base (np.ndaray): fitting curve for baseline
       res (np.ndarray): residual curve (data-fit) for static spectrum (n,)
       edge ({'g', 'l'}): type of edge function, if edge is None then edge function is not
        included in the fitting model

            'g': gaussian type edge function

            'l': lorenzian type edge function
       base_order (int): order of baseline function
                         if base_order is None then baseline is not included in the fitting model
       param_name (np.ndarray): name of parameter
       n_voigt (int): number of voigt component
       n_edge (int): number of edge component
       x (np.ndarray): best parameter
       bounds (sequence of tuple): boundary of each parameter
       c (np.ndarray): best weight of each voigt component and edge of data
       chi2 (float): chi squared value of fitting
       aic (float): Akaike Information Criterion statistic: :math:`N\\log(\\chi^2/N)+2N_{parm}`
       bic (float): Bayesian Information Criterion statistic: :math:`N\\log(\\chi^2/N)+N_{parm}\log(N)`
       red_chi2 (float): total reduced chi squared value of fitting
       nfev (int): total number of function evaluation
       n_param (int): total number of effective parameter
       num_pts (int): total number of data points 
       jac (np.ndarray): jacobian of objective function at optimal point
       cov (np.ndarray): covariance matrix (i.e. inverse of (jac.T @ jac))
       cov_scaled (np.ndarray): scaled covariance matrix (i.e. `red_chi2` * `cov`)
       corr (np.ndarray): parameter correlation matrix
       x_eps (np.ndarray): estimated error of parameter
        (i.e. square root of diagonal element of `conv_scaled`)
       method_glb ({'basinhopping'}): 
        method of global optimization used in fitting process
       message_glb (str): messages from global optimization process
       method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                             minimization (refinement of global optimization solution)
       success_lsq (bool): whether or not local least square optimization is successed
       message_lsq (str): messages from local least square optimization process
       status ({0, -1}): status of optimization process

                     `0` : normal termination

                     `-1` : least square optimization process is failed
      '''

      def __getattr__(self, name):
            try:
                  return self[name]
            except KeyError as e:
                  raise AttributeError(name) from e
      
      __setattr__ = dict.__setitem__
      __delattr__ = dict.__delitem__
      
      def __dir__(self):
            return list(self.keys())
      
      def __str__(self, corr_tol: float = 1e-1):
            '''
            Report StaticResult
            Args:
            corr_tol: parameter correlation greather than `corr_tol` would be reported
            '''
            doc_lst = []
            doc_lst.append("[Model information]")
            doc_lst.append(f"    model : {self['model']}")
            if self['model'] == 'thy':
                  doc_lst.append(f"    policy: {self['policy']}")
            if self['edge'] is not None:
                  doc_lst.append(f"    edge: {self['edge']}")
            if self['base_order'] is not None:
                  doc_lst.append(f"    base_order: {self['base_order']}")
            doc_lst.append(' ')
            doc_lst.append("[Optimization Method]")
            if self['method_glb'] is not None:
                  doc_lst.append(f"    global: {self['method_glb']}")
            doc_lst.append(f"    leastsq: {self['method_lsq']}")
            doc_lst.append(' ')
            doc_lst.append("[Optimization Status]")
            doc_lst.append(f"    nfev: {self['nfev']}")
            doc_lst.append(f"    status: {self['status']}")
            if self['method_glb'] is not None:
                  doc_lst.append(f"    global_opt msg: {self['message_glb']}")
            doc_lst.append(f"    leastsq_opt msg: {self['message_lsq']}")
            doc_lst.append(' ')
            doc_lst.append("[Optimization Results]")
            doc_lst.append(f"    Data points: {self['num_pts']}")
            doc_lst.append(f"    Number of effective parameters: {self['n_param']}")
            doc_lst.append(f"    Degree of Freedom: {self['num_pts']-self['n_param']}")
            doc_lst.append(f"    Chi squared: {self['chi2']: .4f}".rstrip('0').rstrip('.'))
            doc_lst.append(f"    Reduced chi squared: {self['red_chi2']: .4f}".rstrip('0').rstrip('.'))
            doc_lst.append(f"    AIC (Akaike Information Criterion statistic): {self['aic']: .4f}".rstrip('0').rstrip('.'))
            doc_lst.append(f"    BIC (Bayesian Information Criterion statistic): {self['bic']: .4f}".rstrip('0').rstrip('.'))
            doc_lst.append(' ')
            doc_lst.append("[Parameters]")
            for pn, pv, p_err in zip(self['param_name'], self['x'], self['x_eps']):
                  doc_lst.append(f"    {pn}: {pv: .8f} +/- {p_err: .8f} ({100*np.abs(p_err/pv): .2f}%)".rstrip('0').rstrip('.'))
            doc_lst.append(' ')
            doc_lst.append("[Parameter Bound]")
            for pn, pv, bd in zip(self['param_name'], self['x'], self['bounds']):
                  doc_lst.append(f"    {pn}: {bd[0]: .8f}".rstrip('0').rstrip('.')+f" <= {pv: .8f} <= {bd[1]: .8f}".rstrip('0').rstrip('.'))
            doc_lst.append(' ')

            doc_lst.append("[Component Contribution]")
            doc_lst.append(f"    Static spectrum")
            if self['base_order'] is None:
                  coeff_abs = np.abs(self['c'])
                  coeff_contrib = 100*self['c']
            else:
                  coeff_abs = np.abs(self['c'][:-self['base_order']-1])
                  coeff_contrib = 100*self['c'][:-self['base_order']-1]
            coeff_sum = np.sum(coeff_abs)
            coeff_contrib = coeff_contrib/coeff_sum
            for v in range(self['n_voigt']):
                  row = [f"     {self['model']} {v+1}:"]
                  row.append(f'{coeff_contrib[v]: .2f}%')
                  doc_lst.append(' '.join(row))
            if self['edge'] is not None:
                  for e in range(self['n_edge']):
                        row = [f"     {self['edge']} type edge {e+1}:"]
                        row.append(f"{coeff_contrib[self['n_voigt']+e]: .2f}%")
                        doc_lst.append(' '.join(row))
            doc_lst.append(' ')

            doc_lst.append("[Parameter Correlation]")
            doc_lst.append(f"    Parameter Correlations > {corr_tol: .3f}".rstrip('0').rstrip('.') + " are reported.")

            A = np.empty((len(self['x']), len(self['x'])), dtype=object)
            for i in range(len(self['x'])):
                  for j in range(len(self['x'])):
                        A[i,j] = (i,j)
            
            mask = (np.abs(self['corr']) > corr_tol)

            for pair in A[mask]:
                  if pair[0] > pair[1]:
                        tmp_str_lst = [f"    ({self['param_name'][pair[0]]},"]
                        tmp_str_lst.append(f"{self['param_name'][pair[1]]})")
                        tmp_str_lst.append('=')
                        tmp_str_lst.append(f"{self['corr'][pair]: .3f}".rstrip('0').rstrip('.'))
                        doc_lst.append(' '.join(tmp_str_lst))

            return '\n'.join(doc_lst)


def save_StaticResult(result: StaticResult, filename: str):
      '''
      save static fitting result to the h5 file

      Args:
       result: static fitting result
       filename: filename to store result. It will store result to filename.h5

      Returns:
       h5 file which stores result
      '''
      model_key_lst = ['chi2', 'n_param', 'num_pts', 'red_chi2',
      'aic', 'bic', 'nfev', 'method_lsq', 'success_lsq', 'message_lsq', 'status']

      with h5.File(f'{filename}.h5', 'w') as f:
            expt = f.create_group("experiment")
            fit_res = f.create_group("fitting_result")
            if result['model'] == 'thy':
                  thy_stick = f.create_group("theoretical_peaks")
                  for i in range(result['n_voigt']):
                        thy_stick.create_dataset(f'species {i+1}', data=result['thy_peak'][i])
                  fit_res.attrs['policy'] = result['policy']
            expt.create_dataset("energy", data=result['e'])
            expt.create_dataset("intensity", data=result['intensity'])
            expt.create_dataset("error", data=result['eps'])
            fit_res.attrs['model'] = result['model']
            fit_res.attrs['n_voigt'] = result['n_voigt']
            fit_res.attrs['n_edge'] = result['n_edge']
            if result['edge'] is not None:
                  fit_res.attrs['edge'] = result['edge']
            if result['base_order'] is not None:
                  fit_res.attrs['base_order'] = result['base_order']
            else:
                  fit_res.attrs['base_order'] = 'no'
            if result['method_glb'] is not None:
                  fit_res.attrs['method_glb'] = result['method_glb']
                  fit_res.attrs['message_glb'] = result['message_glb']
            else:
                  fit_res.attrs['method_glb'] = 'no'

            for k in model_key_lst:
                  fit_res.attrs[k] = result[k]

            fit_res_curve = fit_res.create_group("fit_curve")
            fit_res_curve.create_dataset("fit", data=result['fit'])
            fit_res_curve.create_dataset("fit_comp", data=result['fit_comp'])
            fit_res_curve.create_dataset("weight", data=result['c'])
            if result['base_order'] is not None:
                  fit_res_curve.create_dataset("base", data=result['base'])
            fit_res_curve.create_dataset("res", data=result['res'])
            fit_res_param = fit_res.create_group("parameter")
            fit_res_param.create_dataset("param_opt", data=result['x'])
            fit_res_param.create_dataset("param_eps", data=result['x_eps'])
            fit_res_param.create_dataset("param_name", data=result['param_name'].astype("S100"), dtype='S100')
            fit_res_param.create_dataset("param_bounds", data=result['bounds'])
            fit_res_param.create_dataset("correlation", data=result['corr'])
            fit_res_mis = fit_res.create_group("miscellaneous")
            fit_res_mis.create_dataset("jac", data=result['jac'])
            fit_res_mis.create_dataset('cov', data=result['cov'])
            fit_res_mis.create_dataset('cov_scaled', data=result['cov_scaled'])
      return

def load_StaticResult(filename: str) -> StaticResult:
      '''
      load static fitting result from h5 file

      Args:
       filename: filename to load result. It will load result to filename.h5
       
      Returns:
       loaded StaticResult instance
      '''
      model_key_lst = ['chi2', 'n_param', 'num_pts', 'red_chi2',
      'aic', 'bic', 'nfev',
      'method_lsq', 'success_lsq', 'message_lsq', 'status']

      result = StaticResult()

      with h5.File(f'{filename}.h5', 'r') as f:
            expt = f['experiment']
            fit_res = f['fitting_result']
            result['e'] = np.atleast_1d(expt['energy'])
            result['intensity'] = np.atleast_1d(expt['intensity'])
            result['eps'] = np.atleast_1d(expt['error'])
            result['model'] = fit_res.attrs['model']
            result['n_voigt'] = fit_res.attrs['n_voigt']
            result['n_edge'] = fit_res.attrs['n_edge']
            if result['model'] == 'thy':
                  thy_stick = f['theoretical_peaks']
                  result['thy_peak'] = np.empty(result['n_voigt'], dtype=object)
                  for i in range(result['n_voigt']):
                        result['thy_peak'][i] = \
                              np.atleast_2d(thy_stick[f'species {i+1}'])
                  result['policy'] = fit_res.attrs['policy']
            if fit_res.attrs['n_edge'] != 0:
                  result['edge'] = fit_res.attrs['edge']
            else:
                  result['edge'] = None
            if fit_res.attrs['base_order'] != 'no':
                  result['base_order'] = fit_res.attrs['base_order']
            else:
                  result['base_order'] = None
            
            if fit_res.attrs['method_glb'] == 'no':
                  result['method_glb'] = None
                  result['message_glb'] = None
            else:
                  result['method_glb'] = fit_res.attrs['method_glb']
                  result['message_glb'] = fit_res.attrs['message_glb']


            for k in model_key_lst:
                  result[k] = fit_res.attrs[k]
            fit_res_curve = fit_res['fit_curve']

            result['fit'] = np.atleast_1d(fit_res_curve['fit'])
            result['fit_comp'] = np.atleast_1d(fit_res_curve['fit_comp'])
            result['c'] = np.atleast_1d(fit_res_curve['weight'])
            if result['base_order'] is not None:
                  result['base'] = np.atleast_1d(fit_res_curve['base'])
            else:
                  result['base'] = None
            result['res'] = np.atleast_1d(fit_res_curve['res'])

            fit_res_param = fit_res['parameter']
            result['x'] = np.atleast_1d(fit_res_param['param_opt'])
            result['x_eps'] = np.atleast_1d(fit_res_param['param_eps'])
            result['param_name'] = np.atleast_1d(fit_res_param['param_name'])
            tmp = np.atleast_2d(fit_res_param['param_bounds'])
            lst = tmp.shape[0]*[None]
            for i in range(tmp.shape[0]):
                  lst[i] = (tmp[i,0], tmp[i, 1])
            result['bounds'] = lst
            result['corr'] = np.atleast_2d(fit_res_param['correlation'])

            fit_res_mis = fit_res['miscellaneous']
            result['jac'] = np.atleast_2d(fit_res_mis['jac'])
            result['cov'] = np.atleast_2d(fit_res_mis['cov'])
            result['cov_scaled'] = np.atleast_2d(fit_res_mis['cov_scaled'])
      return result

def deriv_edge_g_aux(e: np.ndarray, fwhm_G: float) -> np.ndarray:
    '''
    derivative of gaussian type edge

    Args:
     e: energy
     fwhm_G: full width at half maximum
    
    Returns:
     first derivative of gaussian edge function
    '''
    tmp = np.exp(-4*np.log(2)*(e/fwhm_G)**2)/np.sqrt(np.pi)

    return 2*np.sqrt(np.log(2))/fwhm_G*tmp

def dderiv_edge_g_aux(e: np.ndarray, fwhm_G: float) -> np.ndarray:
    '''
    2nd derivative of gaussian type edge

    Args:
     e: energy
     fwhm_G: full width at half maximum
    
    Returns:
     2nd derivative of gaussian edge function
    '''
    tmp = -8*np.log(2)*e*np.exp(-4*np.log(2)*(e/fwhm_G)**2)/(np.sqrt(np.pi)*fwhm_G**2)

    return 2*np.sqrt(np.log(2))/fwhm_G*tmp

def deriv_edge_l_aux(e: np.ndarray, fwhm_L: float) -> np.ndarray:
    '''
    1st derivative of lorenzian type edge

    Args:
     e: energy
    
    Returns:
     first derivative of lorenzian type function
    '''
    tmp = 1/np.pi/(e**2+fwhm_L**2/4)
    return fwhm_L*tmp/2

def dderiv_edge_l_aux(e: np.ndarray, fwhm_L: float) -> np.ndarray:
    '''
    2nd derivative of lorenzian type edge

    Args:
     e: energy
    
    Returns:
     second derivative of lorenzian type function
    '''
    tmp = -e/np.pi/(e**2+fwhm_L**2/4)**2
    return fwhm_L*tmp

def deriv_voigt_aux(e: np.ndarray, fwhm_G: float, fwhm_L: float) -> np.ndarray:
    '''
    1st derivative of voigt profile

    Args:
     e: energy
     fwhm_G: full width at half maximum of gaussian part :math:(2\\sqrt{2\\log(2)}\\sigma)
     fwhm_L: full width at half maximum of lorenzian part :math:(2\\gamma)
    
    Returns:
     first derivative of voigt profile 
    '''
    if fwhm_G < 1e-8:
        tmp = fwhm_L/np.pi/(e**2+fwhm_L**2/4)**2
        return -e*tmp
    
    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    if fwhm_L < 1e-8:
        tmp = np.exp(-(e/sigma)**2/2)/(sigma*np.sqrt(2*np.pi))
        return -e/sigma**2*tmp
    
    z = (e+complex(0,fwhm_L/2))/(sigma*np.sqrt(2))
    return -(z*wofz(z)).real/(sigma**2*np.sqrt(np.pi))

def dderiv_voigt_aux(e: np.ndarray, fwhm_G: float, fwhm_L: float) -> np.ndarray:
    '''
    2nd derivative of voigt profile

    Args:
     e: energy
     fwhm_G: full width at half maximum of gaussian part :math:(2\\sqrt{2\\log(2)}\\sigma)
     fwhm_L: full width at half maximum of lorenzian part :math:(2\\gamma)
    
    Returns:
     second derivative of voigt profile 
    '''

    if fwhm_G < 1e-8:
        tmp = fwhm_L/np.pi/(e**2+fwhm_L**2/4)**2
        return (4*e**2/(e**2+fwhm_L**2/4)-1)*tmp
    
    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    if fwhm_L < 1e-8:
        tmp = np.exp(-(e/sigma)**2/2)/(sigma*np.sqrt(2*np.pi))
        return 1/sigma**2*((e/sigma)**2-1)*tmp
    
    z = (e+complex(0,fwhm_L/2))/(sigma*np.sqrt(2))
    f = wofz(z)/(sigma*np.sqrt(2*np.pi))
    return 1/sigma**4*\
      ((e**2-fwhm_L**2/4-sigma**2)*f.real - e*fwhm_L*f.imag + fwhm_L/(2*np.pi))

def voigt_thy_aux(e: np.ndarray, thy_peak: np.ndarray,
                  fwhm_G: float, fwhm_L: float,
                  deriv_order: int = 0) -> np.ndarray:

    '''
    Calculates derivative of normalized 
    voigt broadened theoretically calculated lineshape spectrum

    Args:
        e: energy 
        thy_peak: theoretical calculated peak position and intensity
        fwhm_G: full width at half maximum of gaussian shape 
        fwhm_L: full width at half maximum of lorenzian shape 
        deriv_order({0, 1, 2}): order of derivative

    Returns:
      derivative of normalized voigt broadened theoritical lineshape spectrum 
    '''

    v_matrix = np.empty((e.size, thy_peak.shape[0]))

    if deriv_order == 0:
      for i in range(thy_peak.shape[0]):
            v_matrix[:, i] = voigt(e-thy_peak[i, 0], fwhm_G, fwhm_L)
    elif deriv_order == 1:
      for i in range(thy_peak.shape[0]):
            v_matrix[:, i] = deriv_voigt_aux(e-thy_peak[i, 0], fwhm_G, fwhm_L)
    else:
      for i in range(thy_peak.shape[0]):
            v_matrix[:, i] = dderiv_voigt_aux(e-thy_peak[i, 0], fwhm_G, fwhm_L)

    broadened_theory = v_matrix @ thy_peak[:, 1].reshape((thy_peak.shape[0], 1))

    return broadened_theory.flatten()

def static_spectrum(e: np.ndarray, 
                    result: StaticResult, deriv_order: int = 0) -> Callable:
      '''
      Evaluates static spectrum from static spectrum
      fitting result.

      Args:
       e: energy at which evaulate static spectrum
       result: static spectrum fitting result
       deriv_order ({0, 1, 2}): order of derivative. [default: 0]
      
      Returns:
       Evaluated static spectrum
      
      Note:
       Baseline feature is removed.
      '''

      tmp = np.zeros(e.size)

      if result['model'] == 'voigt' and result['n_voigt'] > 0:
            e0_voigt = result['x'][:result['n_voigt']]
            fwhm_G_voigt = result['x'][result['n_voigt']:2*result['n_voigt']]
            fwhm_L_voigt = result['x'][2*result['n_voigt']:3*result['n_voigt']]
            V = np.empty((result['n_voigt'], e.size))
            if deriv_order == 0:
                  for i in range(result['n_voigt']):
                        V[i, :] = voigt(e-e0_voigt[i], fwhm_G_voigt[i], fwhm_L_voigt[i])
            elif deriv_order == 1:
                  for i in range(result['n_voigt']):
                        V[i, :] = deriv_voigt_aux(e-e0_voigt[i], fwhm_G_voigt[i],
                        fwhm_L_voigt[i])
            else:
                  for i in range(result['n_voigt']):
                        V[i, :] = dderiv_voigt_aux(e-e0_voigt[i], fwhm_G_voigt[i],
                        fwhm_L_voigt[i])
            tmp = (result['c'][:result['n_voigt']]@V).flatten()
      else:
            if result['policy'] == 'shift':
                  for i in range(result['n_voigt']):
                        peak_tmp = result['thy_peak'][i].copy()
                        peak_tmp[:, 0] = peak_tmp[:, 0]+result['x'][2+i]
                        tmp = tmp + result['c'][i]*voigt_thy_aux(e, peak_tmp,
                        result['x'][0], result['x'][1], deriv_order)
            elif result['policy'] == 'scale':
                  for i in range(result['n_voigt']):
                        peak_tmp = result['thy_peak'][i].copy()
                        peak_tmp[:, 0] = peak_tmp[:, 0]*result['x'][2+i]
                        tmp = tmp + result['c'][i]*voigt_thy_aux(e, peak_tmp,
                        result['x'][0], result['x'][1], deriv_order)
            else:
                  for i in range(result['n_voigt']):
                        peak_tmp = result['thy_peak'][i].copy()
                        peak_tmp[:, 0] = result['x'][2+result['n_voigt']+i]*peak_tmp[:, 0]+\
                              result['x'][2+i]
                        tmp = tmp + result['c'][i]*voigt_thy_aux(e, peak_tmp,
                        result['x'][0], result['x'][1], deriv_order)
      
      if result['model'] == 'voigt':
            param_edge_start = 3*result['n_voigt']
      else:
            if result['policy'] in ['scale', 'shift']:
                  param_edge_start = 2+result['n_voigt']
            else:
                  param_edge_start = 2+2*result['n_voigt']

      if result['edge'] == 'g':
            if deriv_order == 0:
                  for i in range(result['n_edge']):
                        tmp = tmp + result['c'][result['n_voigt']+i]*\
                        edge_gaussian(e-result['x'][param_edge_start+i],
                        result['x'][param_edge_start+result['n_edge']+i])
            elif deriv_order == 1:
                  for i in range(result['n_edge']):
                        tmp = tmp + result['c'][result['n_voigt']+i]*\
                        deriv_edge_g_aux(e-result['x'][param_edge_start+i],
                        result['x'][param_edge_start+result['n_edge']+i])
            else:
                  for i in range(result['n_edge']):
                        tmp = tmp + result['c'][result['n_voigt']+i]*\
                        dderiv_edge_g_aux(e-result['x'][param_edge_start+i],
                        result['x'][param_edge_start+result['n_edge']+i])
      elif result['edge'] == 'l':
            if deriv_order == 0:
                  for i in range(result['n_edge']):
                        tmp = tmp + result['c'][result['n_voigt']+i]*\
                        edge_lorenzian(e-result['x'][param_edge_start+i],
                        result['x'][param_edge_start+result['n_edge']+i])
            elif deriv_order == 1:
                  for i in range(result['n_edge']):
                        tmp = tmp + result['c'][result['n_voigt']+i]*\
                        deriv_edge_l_aux(e-result['x'][param_edge_start+i],
                        result['x'][param_edge_start+result['n_edge']+i])
            else:
                  for i in range(result['n_edge']):
                        tmp = tmp + result['c'][result['n_voigt']+i]*\
                        dderiv_edge_l_aux(e-result['x'][param_edge_start+i],
                        result['x'][param_edge_start+result['n_edge']+i])
      return tmp



# calc_dads.py
# Wrapper script for calc_dads()
# Date: 2022. 7. 25.
# Author: pistack
# Email: pistack@yonsei.ac.kr

import os
import sys

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack.tools._calc_dads import calc_dads

if __name__ == "__main__":
    calc_dads()



import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import exp_conv, fact_anal_exp_conv
from TRXASprefitpack import dmp_osc_conv, fact_anal_dmp_osc_conv
from TRXASprefitpack import sum_exp_dmp_osc_conv, fact_anal_sum_exp_dmp_osc_conv

class TestFactAnal(unittest.TestCase):

    def test_fact_anal_exp_conv_1(self):
        '''
        Test factor analysis routine (uniform noise level)
        '''
        fwhm = 0.15
        c_ref = np.array([1, -0.5, -0.25])
        tau = np.array([0.3, 10])
        t = np.hstack((np.arange(-1, 1, 0.05), np.linspace(1, 100, 99)))
        model = exp_conv(t, fwhm, tau, c_ref, True, irf='g')
        noise = np.random.normal(0, np.max(np.abs(model))/10, model.size)
        eps = np.max(np.abs(model))/10*np.ones_like(noise)
        expt = model+noise
        c_tst = fact_anal_exp_conv(t, fwhm, tau, True, irf='g', intensity=expt, eps=eps)
        cond = np.linalg.norm(c_ref-c_tst)/(np.linalg.norm(c_ref)) < 1e-1
        self.assertEqual(cond, True)

    def test_fact_anal_exp_conv_2(self):
        '''
        Test factor analysis routine (heterogeneous noise)
        '''
        fwhm = 0.15
        c_ref = np.array([1, -0.5, -0.25])
        tau = np.array([0.3, 10])
        t = np.hstack((np.arange(-1, 1, 0.05), np.linspace(1, 100, 99)))
        model = exp_conv(t, fwhm, tau, c_ref, True, irf='g')
        eps = np.max(np.abs(model))/10*np.random.normal(1, 0.1, model.size)
        noise = np.empty_like(eps)
        for i in range(noise.size):
            noise[i] = np.random.normal(0, eps[i])
        expt = model+noise
        c_tst = fact_anal_exp_conv(t, fwhm, tau, True, irf='g', intensity=expt, eps=eps)
        cond = np.linalg.norm(c_ref-c_tst)/(np.linalg.norm(c_ref)) < 1e-1
        self.assertEqual(cond, True)

    def test_fact_anal_dmp_osc_conv(self):
        fwhm = 0.15
        c_ref = np.array([1, 0.5])
        tau = np.array([0.3, 1])
        period = np.array([0.5, 1])
        phase = np.array([np.pi/3, -np.pi/4])
        t = np.hstack((np.arange(-1, 1, 0.02), np.linspace(1, 100, 99)))
        model = dmp_osc_conv(t, fwhm, tau, period, phase, c_ref, irf='g')
        noise = np.random.normal(0, np.max(np.abs(model))/100, model.size)
        eps = np.max(np.abs(model))/100*np.ones_like(noise)
        expt = model+noise
        phase_tst, c_tst = fact_anal_dmp_osc_conv(t, fwhm, tau, period, irf='g', intensity=expt, eps=eps)
        cond = np.linalg.norm(c_ref-c_tst)/(np.linalg.norm(c_ref)) < 1e-2
        cond_phase = np.allclose(phase, phase_tst, rtol=1e-2)
        self.assertEqual((cond, cond_phase), (True, True))

    def test_fact_anal_sum_exp_dmp_osc_conv(self):
        fwhm = 0.15
        c_ref_decay = np.array([1, -0.5, -0.25])
        c_ref_osc = np.array([0.15, 0.075])
        tau = np.array([0.5, 10])
        tau_osc = np.array([0.3, 1])
        period_osc = np.array([0.5, 1])
        phase_osc = np.array([np.pi/3, -np.pi/4])
        t = np.hstack((np.arange(-1, 1, 0.02), np.linspace(1, 100, 99)))
        model = sum_exp_dmp_osc_conv(t, fwhm, tau, tau_osc, period_osc, phase_osc, 
        c_ref_decay, c_ref_osc, base=True, irf='g')
        noise = np.random.normal(0, np.max(np.abs(model))/100, model.size)
        eps = np.max(np.abs(model))/100*np.ones_like(noise)
        expt = model+noise
        c_tst_decay, phase_tst_osc, c_tst_osc = \
            fact_anal_sum_exp_dmp_osc_conv(t, fwhm, tau, tau_osc, period_osc, base=True,
            irf='g', intensity=expt, eps=eps)
        cond_decay = np.linalg.norm(c_ref_decay-c_tst_decay)/(np.linalg.norm(c_ref_decay)) < 1e-2
        cond_osc = np.linalg.norm(c_ref_osc-c_tst_osc)/(np.linalg.norm(c_ref_osc)) < 1e-1
        cond_phase_osc = np.allclose(phase_osc, phase_tst_osc, rtol=1e-1)
        self.assertEqual((cond_decay, cond_phase_osc, cond_osc), (True, True, True))


if __name__ == '__main__':
    unittest.main()





'''
ads:
submodule for driver routine of associated difference spectrum

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Tuple 
import numpy as np
from scipy.linalg import svd
from ..mathfun.A_matrix import make_A_matrix_exp
from ..mathfun.rate_eq import compute_signal_irf

def dads(escan_time: np.ndarray, fwhm: float, tau: np.ndarray, base: Optional[bool] = True,
irf: Optional[str] = 'g', eta: Optional[float] = None,
intensity: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculate decay associated difference spectrum from experimental energy scan data

    Args:
      escan_time: time delay for each energy scan data
      fwhm: full width at half maximum of instrumental response function
      tau: life time for each component
      base: whether or not include baseline [default: True]
      irf: shape of instrumental response function [default: g]

           * 'g': normalized gaussian distribution,
           * 'c': normalized cauchy distribution,
           * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`

      eta: mixing parameter for pseudo voigt profile
           (only needed for pseudo voigt profile)
      intensity: intensity of energy scan dataset
      eps: standard error of dataset
    
    Returns:
     Tuple of calculated decay associated difference spectrum of each component, estimated error and
     retrieved energy scan intensity from dads and decay components
    
    Note:
     To calculate decay associated difference spectrum of n component exponential decay, you should measure at least n+1
     energy scan
    '''
    # initialization
    if base:
      c = np.empty((tau.size+1, intensity.shape[0]))
      dof = escan_time.size - (tau.size+1)
    else:
      c = np.empty((tau.size, intensity.shape[0]))
      dof = escan_time.size - (tau.size)

    if eps is None:
      eps = np.ones_like(intensity)
    
    c_eps = np.empty_like(c)
    
    A = make_A_matrix_exp(escan_time, fwhm, tau, base, irf, eta)
    data_scaled = intensity/eps

    # evaluates dads
    cond = 1e-2
    for i in range(intensity.shape[0]):
      A_scaled = np.einsum('j,ij->ij', 1/eps[i,:], A)
      U, s, Vh = svd(A_scaled.T, full_matrices= False)
      mask = s > cond*s[0]
      U_turn = U[:,mask]; s_turn = s[mask]; Vh_turn = Vh[mask, :]
      c[:,i] = np.einsum('j,ij->ij', 1/s_turn, Vh_turn.T) @ (U_turn.T @ data_scaled[i,:])
      res = data_scaled[i,:] - (c[:,i] @ A_scaled)
      if dof != 0:
        red_chi2 = np.sum(res**2)/dof
        cov = Vh_turn.T @ np.einsum('i,ij->ij', 1/s_turn**2, Vh_turn)
        c_eps[:,i] = np.sqrt(red_chi2*np.diag(cov))


    return c, c_eps, c.T @ A

def sads(escan_time: np.ndarray, fwhm: float, eigval: np.ndarray, V: np.ndarray, c: np.ndarray,
exclude: Optional[str] = None, irf: Optional[str] = 'g', eta: Optional[float] = None,
intensity: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculate species associated difference spectrum from experimental energy scan data

    Args:
      escan_time: time delay for each energy scan data
      fwhm: full width at half maximum of instrumental response function
      eigval: eigenvalue of rate equation matrix 
      V: eigenvector of rate equation matrix 
      c: coefficient to match initial condition of rate equation
      exclude: exclude either 'first' or 'last' element or both 'first' and 'last' element.

               * 'first' : exclude first element
               * 'last' : exclude last element
               * 'first_and_last' : exclude both first and last element  
               * None : Do not exclude any element [default]

      irf: shape of instrumental response function [default: g]

           * 'g': normalized gaussian distribution,
           * 'c': normalized cauchy distribution,
           * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
           
      eta: mixing parameter for pseudo voigt profile (only needed for pseudo voigt profile)
      intensity: intensity of energy scan dataset
      eps: standard error of data 
    
    Returns:
     Tuple of calculated species associated difference spectrum of each component, estimated error and
     retrieved intensity of energy scan from sads and model excited state components
    
    Note:
     1. eigval, V, c should be obtained from solve_model
     2. To calculate species associated difference spectrum of n excited state species, you should measure at least n+1 energy scan
     3. Difference spectrum of ground state is zero, so ground state species should be excluded from rate equation or via exclude option.
    '''
    # initialization
    if exclude is None:
      abs = np.empty((eigval.size, intensity.shape[0]))
      dof = escan_time.size - eigval.size
    elif exclude in ['first', 'last']:
      abs = np.empty((eigval.size-1, intensity.shape[0]))
      dof = escan_time.size - (eigval.size-1)
    else:
      abs = np.empty((eigval.size-2, intensity.shape[0]))
      dof = escan_time.size - (eigval.size-2)

    if eps is None:
      eps = np.ones_like(intensity)
    
    abs_eps = np.empty_like(abs)
    
    A = compute_signal_irf(escan_time, eigval, V, c, fwhm, irf, eta)
    if exclude == 'first':
      B = A[1:, :]
    elif exclude == 'last':
      B = A[:-1, :]
    elif exclude == 'first_and_last':
      B = A[1:-1, :]
    else:
      B = A

    data_scaled = intensity/eps

    # evaluates sads
    cond = 1e-2
    for i in range(intensity.shape[0]):
      A_scaled = np.einsum('j,ij->ij', 1/eps[i,:], B)
      U, s, Vh = svd(A_scaled.T, full_matrices= False)
      mask = s > cond*s[0]
      U_turn = U[:,mask]; s_turn = s[mask]; Vh_turn = Vh[mask, :]
      cov = Vh_turn.T @ np.einsum('i,ij->ij', 1/s_turn**2, Vh_turn)
      abs[:,i] = np.einsum('j,ij->ij', 1/s_turn, Vh_turn.T) @ (U_turn.T @ data_scaled[i,:])
      res = data_scaled[i,:] - (abs[:,i] @ A_scaled)
      if dof != 0:
        red_chi2 = np.sum(res**2)/dof
        abs_eps[:,i] = np.sqrt(red_chi2*np.diag(cov))


    return abs, abs_eps, abs.T @ B


import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack.mathfun.deriv_tst import test_num_deriv
from TRXASprefitpack import calc_eta, deriv_eta
from TRXASprefitpack import calc_fwhm, deriv_fwhm
from TRXASprefitpack import pvoigt_irf
from TRXASprefitpack import voigt

class TestPvoigtIRF(unittest.TestCase):
    def test_pvoigt_irf_1(self):
        '''
         Test pseudo voigt approximation (fwhm_L = 3 fwhm_G)
        '''
        t = np.linspace(-1, 1, 201)
        fwhm_G = 0.1; fwhm_L = 0.3
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        ref = voigt(t, fwhm_G, fwhm_L)
        tst = pvoigt_irf(t, fwhm, eta)
        result = np.max(np.abs(tst-ref))/np.max(ref) < 2e-2
        self.assertEqual(result, True)

    def test_pvoigt_irf_2(self):
        '''
         Test pseudo voigt approximation (fwhm_L = 1.5 fwhm_G)
        '''
        t = np.linspace(-1, 1, 201)
        fwhm_G = 0.1; fwhm_L = 0.15
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        ref = voigt(t, fwhm_G, fwhm_L)
        tst = pvoigt_irf(t, fwhm, eta)
        result = np.max(np.abs(tst-ref))/np.max(ref) < 2e-2
        self.assertEqual(result, True)

    def test_pvoigt_irf_3(self):
        '''
         Test pseudo voigt approximation (fwhm_L = 1 fwhm_G)
        '''
        t = np.linspace(-1, 1, 201)
        fwhm_G = 0.1; fwhm_L = 0.1
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        ref = voigt(t, fwhm_G, fwhm_L)
        tst = pvoigt_irf(t, fwhm, eta)
        result = np.max(np.abs(tst-ref))/np.max(ref) < 2e-2
        self.assertEqual(result, True)

    def test_pvoigt_irf_4(self):
        '''
         Test pseudo voigt approximation (fwhm_G = 1.5 fwhm_L)
        '''
        t = np.linspace(-1, 1, 201)
        fwhm_G = 0.15; fwhm_L = 0.1
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        ref = voigt(t, fwhm_G, fwhm_L)
        tst = pvoigt_irf(t, fwhm, eta)
        result = np.max(np.abs(tst-ref))/np.max(ref) < 2e-2
        self.assertEqual(result, True)

    def test_pvoigt_irf_5(self):
        '''
         Test pseudo voigt approximation (fwhm_G = 3 fwhm_L)
        '''
        t = np.linspace(-1, 1, 201)
        fwhm_G = 0.3; fwhm_L = 0.1
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        ref = voigt(t, fwhm_G, fwhm_L)
        tst = pvoigt_irf(t, fwhm, eta)
        result = np.max(np.abs(tst-ref))/np.max(ref) < 2e-2
        self.assertEqual(result, True)
    
    def test_deriv_eta_1(self):
        '''
         Test gradient of mixing parameter eta (fwhm_L = 3 fwhm_G)
        '''
        fwhm_G = 0.1; fwhm_L = 0.3
        d_G, d_L = deriv_eta(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_eta, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_eta_2(self):
        '''
         Test gradient of mixing parameter eta (fwhm_L = 1.5 fwhm_G)
        '''
        fwhm_G = 0.1; fwhm_L = 0.15
        d_G, d_L = deriv_eta(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_eta, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_eta_3(self):
        '''
         Test gradient of mixing parameter eta (fwhm_L = fwhm_G)
        '''
        fwhm_G = 0.1; fwhm_L = 0.1
        d_G, d_L = deriv_eta(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_eta, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_eta_4(self):
        '''
         Test gradient of mixing parameter eta (fwhm_G = 1.5 fwhm_L)
        '''
        fwhm_G = 0.15; fwhm_L = 0.1
        d_G, d_L = deriv_eta(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_eta, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_eta_5(self):
        '''
         Test gradient of mixing parameter eta (fwhm_G = 3 fwhm_L)
        '''
        fwhm_G = 0.3; fwhm_L = 0.1
        d_G, d_L = deriv_eta(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_eta, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_fwhm_1(self):
        '''
         Test gradient of unifrom fwhm parameter (fwhm_L = 3 fwhm_G)
        '''
        fwhm_G = 0.1; fwhm_L = 0.3
        d_G, d_L = deriv_fwhm(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_fwhm, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_fwhm_2(self):
        '''
         Test gradient of unifrom fwhm parameter (fwhm_L = 1.5 fwhm_G)
        '''
        fwhm_G = 0.1; fwhm_L = 0.15
        d_G, d_L = deriv_fwhm(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_fwhm, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_fwhm_3(self):
        '''
         Test gradient of unifrom fwhm parameter (fwhm_L = fwhm_G)
        '''
        fwhm_G = 0.1; fwhm_L = 0.1
        d_G, d_L = deriv_fwhm(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_fwhm, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_fwhm_4(self):
        '''
         Test gradient of uniform fwhm parameter (fwhm_G = 1.5 fwhm_L)
        '''
        fwhm_G = 0.15; fwhm_L = 0.1
        d_G, d_L = deriv_fwhm(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_fwhm, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

    def test_deriv_fwhm_5(self):
        '''
         Test gradient of unifrom fwhm parameter (fwhm_G = 3 fwhm_L)
        '''
        fwhm_G = 0.3; fwhm_L = 0.1
        d_G, d_L = deriv_fwhm(fwhm_G, fwhm_L)
        
        result = np.allclose(np.array([d_G, d_L]), 
        test_num_deriv(calc_fwhm, fwhm_G, fwhm_L))
        self.assertEqual(result, True)

if __name__ == '__main__':
    unittest.main()



# fit static
# fitting static spectrum

import os
from pathlib import Path
import argparse
import numpy as np
import matplotlib.pyplot as plt
from ..driver import StaticResult, save_StaticResult
from ..driver import fit_static_voigt, fit_static_thy


def save_StaticResult_txt(result: StaticResult, dirname: str):
      '''
      save static fitting result to the text file

      Args:
       result: static fitting result
       dirname: name of the directory in which text files for fitting result are saved.
       e: energy range of static spectrum
       eps: estimated error of static spectrum
      
      Returns:
       `fit_summary.txt`: Summary for the fitting result
       `weight.txt`: Weight of each voigt and edge component
       `fit.txt`: fitting, each voigt, edge and baseline curve for static spectrum
       `res.txt`: residual (fit-data) curve for static spectrum

      Note:
       If `dirname` directory is not exists, it creates `dirname` directory.
      '''
      if not (Path.cwd()/dirname).exists():
            os.mkdir(dirname)
      
      with open(f'{dirname}/fit_summary.txt', 'w') as f:
            f.write(str(result))

      tot_comp = result['n_voigt']
      if result['edge'] is not None:
            tot_comp = tot_comp+result['n_edge']
      if result['base_order'] is not None:
            tot_comp = tot_comp+1
      coeff_fmt = ['%.8e']
      fit_fmt = (2+tot_comp)*['%.8e']
      fit_header_lst = ['energy', 'fit']
      for i in range(result['n_voigt']):
            fit_header_lst.append(f'voigt_{i}')
      if result['edge'] is not None:
            for i in range(result['n_edge']):
                fit_header_lst.append(f"{result['edge']}_type_edge {i+1}")
      if result['base'] is not None:
            fit_header_lst.append("base")
            fit_save = np.vstack((result['e'], result['fit'], result['fit_comp'], result['base'])).T
      else:
            fit_save = np.vstack((result['e'], result['fit'], result['fit_comp'])).T
      res_save = np.vstack((result['e'], result['res'], result['eps'])).T
      np.savetxt(f'{dirname}/res.txt', res_save, fmt=['%.8e', '%.8e', '%.8e'], 
      header=f'energy \t res \t eps')
      fit_header = '\t'.join(fit_header_lst)
      coeff_header = 'static'

      np.savetxt(f'{dirname}/weight.txt', result['c'], fmt=coeff_fmt, header=coeff_header)
      np.savetxt(f'{dirname}/fit.txt', fit_save, fmt=fit_fmt, header=fit_header)
      
      return

def plot_StaticResult(result: StaticResult):
      '''
      plot static fitting Result

      Args:
       result: static fitting result
       name_of_dset: name of each dataset
       x_min: minimum x range
       x_max: maximum x range
       save_fig: prefix of saved png plots. If `save_fig` is `None`, plots are displayed istead of being saved.
       e: scan range of data
       data: static spectrum (it should not contain energy scan range)
       eps: estimated errors of static spectrum
      '''
      
      fig = plt.figure(0)
      title = 'Static Spectrum'
      subtitle = f"Chi squared: {result['red_chi2']: .2f}"
      plt.suptitle(title)
      sub1 = fig.add_subplot(211)
      sub1.set_title(subtitle)
      sub1.errorbar(result['e'], result['intensity'], result['eps'], 
      marker='o', mfc='none', label=f'expt {title}', linestyle='none')
      sub1.plot(result['e'], result['fit'], label=f'fit {title}')
      for i in range(result['n_voigt']):
            sub1.plot(result['e'], result['fit_comp'][i, :], label=f'{i+1}th voigt component', linestyle='dashed')
      if result['edge'] is not None:
            for i in range(result['n_edge']):
                sub1.plot(result['e'], result['fit_comp'][result['n_voigt']+i, :], 
                label=f"{result['edge']} type edge {i+1}", linestyle='dashed')
      if result['base_order'] is not None:
        sub1.plot(result['e'], result['base'], label=f"base [order {result['base_order']}]", linestyle='dashed')
      sub1.legend()
      sub2 = fig.add_subplot(212)
      sub2.errorbar(result['e'], result['res'], result['eps'], 
      marker='o', mfc='none', label=f'res {title}', linestyle='none')
      sub2.legend()
      plt.show()
      return

description = '''
fit static: fitting static spectrum with 
 'voigt': sum of voigt component
  'thy' : theoretically calculated line spectrum broadened by voigt function
It also include edge and polynomial type baseline feature.
'''

epilog = ''' 
*Note
 If fwhm_G of voigt component is zero then this voigt component is treated as lorenzian
 If fwhm_L of voigt component is zero then this voigt component is treated as gaussian
'''

do_glb_help='''
Whether or not use Global optimization Method.
When do_glb option is set then `basinhopping` global optimization algorithm will be used.
'''

edge_help = '''
Type of edge function if not set, edge is not included.
 'g': gaussian type edge function
 'l': lorenzian type edge function
'''

mode_help = '''
Mode of static spectrum fitting
 'voigt': fitting with sum of voigt componenty
 'thy': fitting with voigt broadend thoretical spectrum
'''

policy_help = '''
Policy to match discrepency between experimental data and theoretical spectrum.
 'shift': constant shift peak position
 'scale': constant scale peak position
 'both': shift and scale peak position
'''


def fit_static():

    tmp = argparse.RawDescriptionHelpFormatter
    parse = argparse.ArgumentParser(formatter_class=tmp,
                                    description=description,
                                    epilog=epilog)
    parse.add_argument('filename', help='filename for experimental spectrum')
    parse.add_argument('--mode', type=str, choices=['voigt', 'thy'],
    help = mode_help)
    parse.add_argument('--e0_voigt', type=float, nargs='*',
    help='peak position of each voigt component')
    parse.add_argument('--fwhm_G_voigt', type=float, nargs='*',
                        help='full width at half maximum for gaussian shape ' +
                        'It would be not used when you set lorenzian line shape')
    parse.add_argument('--fwhm_L_voigt', type=float, nargs='*',
                        help='full width at half maximum for lorenzian shape ' +
                        'It would be not used when you use gaussian line shape')
    parse.add_argument('--thy_file', type=str, nargs='*',
    help='filenames which store thoretical peak position and intensity.')
    parse.add_argument('--fwhm_G_thy', type=float, default=0, help='gaussian part of uniform' +
    ' broadening parameter for theoretical line shape spectrum')
    parse.add_argument('--fwhm_L_thy', type=float, default=0, help='lorenzian part of uniform' +
    ' broadening parameter for theoretical line shape spectrum')
    parse.add_argument('--policy', choices=['shift', 'scale', 'both'], help=policy_help)
    parse.add_argument('--peak_scale', type=float, nargs='*', help='inital peak position scale parameter')
    parse.add_argument('--peak_shift', type=float, nargs='*', help='inital peak position shift parameter')
    parse.add_argument('--edge', type=str, choices=['g', 'l'],
    help=edge_help)

    parse.add_argument('--e0_edge', type=float, nargs='*',
    help='edge position')
    parse.add_argument('--fwhm_edge', type=float, nargs='*',
    help='full width at half maximum parameter of edge')
    parse.add_argument('--base_order', type=int,
    help ='Order of polynomial to correct baseline feature. If it is not set then baseline is not corrected')
    parse.add_argument('-o', '--outdir', default='out', help='directory to store output file')
    parse.add_argument('--do_glb', action='store_true', help=do_glb_help)

    args = parse.parse_args()

    filename = args.filename
    if args.mode == 'voigt' and args.e0_voigt is None:
        e0_init = None
        fwhm_G_init = None
        fwhm_L_init = None
    elif args.mode == 'voigt' and args.e0_voigt is not None:
        e0_init = np.array(args.e0_voigt)
        if args.fwhm_G_voigt is None:
            fwhm_G_init = np.zeros_like(e0_init)
        else:
            fwhm_G_init = np.array(args.fwhm_G_voigt)
        if args.fwhm_L_voigt is None:
            fwhm_L_init = np.zeros_like(e0_init)
        else:
            fwhm_L_init = np.array(args.fwhm_L_voigt)
        if args.fwhm_G_voigt is None and args.fwhm_L_voigt is None:
            raise Exception("Please set both initial fwhm_G and fwhm_L for each voigt component")
        if fwhm_G_init.size != fwhm_L_init.size:
            raise Exception("The number of initial fwhm_G and fwhm_L parameter should be same")
    elif args.mode == 'thy':
        fwhm_G_init = args.fwhm_G_thy
        fwhm_L_init = args.fwhm_L_thy
        thy_peak = np.empty(len(args.thy_file), dtype=object)
        for i in range(thy_peak.size):
            thy_peak[i] = np.genfromtxt(args.thy_file[i])[:,:2]
        if args.policy is None:
            raise Exception("Please set policy to solve descrepency between theoretical and experimental spectrum.")
        if args.policy in ['shift', 'both'] and args.peak_shift is None:
            raise Exception(f"Your policy is {args.policy}, please set initial peak_shift parameter.")
        if args.policy in ['scale', 'both'] and args.peak_scale is None:
            raise Exception(f"Your policy is {args.policy}, please set peak_scale parameter.")   
    
    edge = args.edge
    e0_edge_init = np.array(args.e0_edge)
    fwhm_edge_init = np.array(args.fwhm_edge)
    base_order = args.base_order
    outdir = args.outdir

    tmp = np.genfromtxt(filename)
    e = tmp[:, 0]
    intensity = tmp[:, 1]
    if tmp.shape[1] == 2:
        eps = np.max(np.abs(intensity))/1000*np.ones_like(e)
    else:
        eps = tmp[:, 2]
    
    if args.mode == 'voigt':
        result = fit_static_voigt(e0_init, fwhm_G_init, fwhm_L_init, edge, e0_edge_init, fwhm_edge_init,
        base_order, args.do_glb, e=e, intensity=intensity, eps=eps)
    
    elif args.mode == 'thy':
        result = fit_static_thy(thy_peak, fwhm_G_init, fwhm_L_init, args.policy, args.peak_shift, args.peak_scale,
        edge, e0_edge_init, fwhm_edge_init,
        base_order, args.do_glb, e=e, intensity=intensity, eps=eps)

    save_StaticResult_txt(result, outdir)
    save_StaticResult(result, outdir)
    print(result)
    plot_StaticResult(result)

    return



import os
import sys
import unittest
import numpy as np
from scipy.signal import convolve

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import calc_eta
from TRXASprefitpack import calc_fwhm
from TRXASprefitpack import gau_irf, cauchy_irf
from TRXASprefitpack import voigt
from TRXASprefitpack import exp_conv_pvoigt, exp_conv_gau, exp_conv_cauchy

def decay(t, k):
    return np.exp(-k*t)*np.heaviside(t, 1)

class TestExpConvIRF(unittest.TestCase):

    def test_exp_conv_gau(self):
        fwhm_G = 0.15; tau = 0.3
        t = np.linspace(-2, 2, 2000)
        t_sample = np.hstack((np.arange(-1, -0.5, 0.1),
        np.arange(-0.5, 0.5, 0.05), np.linspace(0.5, 1, 6)))
        sample_idx = np.searchsorted(t, t_sample)
        gau_ref = gau_irf(t, fwhm_G)
        decay_ref = decay(t, 1/tau)
        ref = convolve(gau_ref, decay_ref, mode='same')*4/2000
        tst = exp_conv_gau(t_sample, fwhm_G, 1/tau)
        cond = np.max(np.abs(ref[sample_idx]-tst))/np.max(ref) < 1e-3
        self.assertEqual(cond, True)


    def test_exp_conv_cauchy(self):
        fwhm_L = 0.15; tau = 0.3
        t = np.linspace(-2, 2, 2000)
        t_sample = np.hstack((np.arange(-1, -0.5, 0.1),
        np.arange(-0.5, 0.5, 0.05), np.linspace(0.5, 1, 6)))
        sample_idx = np.searchsorted(t, t_sample)
        cauchy_ref = cauchy_irf(t, fwhm_L)
        decay_ref = decay(t, 1/tau)
        ref = convolve(cauchy_ref, decay_ref, mode='same')*4/2000
        tst = exp_conv_cauchy(t_sample, fwhm_L, 1/tau)
        cond = np.max(np.abs(ref[sample_idx]-tst))/np.max(ref) < 1e-3
        self.assertEqual(cond, True)

    def test_exp_conv_pvoigt(self):
        fwhm_G = 0.15; fwhm_L = 0.10; tau = 0.3
        fwhm = calc_fwhm(fwhm_G, fwhm_L)
        eta = calc_eta(fwhm_G, fwhm_L)
        t = np.linspace(-2, 2, 2000)
        t_sample = np.hstack((np.arange(-1, -0.5, 0.1),
        np.arange(-0.5, 0.5, 0.05), np.linspace(0.5, 1, 6)))
        sample_idx = np.searchsorted(t, t_sample)
        voigt_ref = voigt(t, fwhm_G, fwhm_L)
        decay_ref = decay(t, 1/tau)
        ref = convolve(voigt_ref, decay_ref, mode='same')*4/2000
        tst = exp_conv_pvoigt(t_sample, fwhm, eta, 1/tau)
        cond = np.max(np.abs(ref[sample_idx]-tst))/np.max(ref) < 1e-2
        self.assertEqual(cond, True)

if __name__ == '__main__':
    unittest.main()




'''
anal_fit:
submodule for 
1. comparing two different fitting model
2. calculating confidence interval of parameter
based on f_test

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
import numpy as np
from scipy.stats import f, norm
from scipy.optimize import brenth, minimize
from ..res import res_grad_decay, res_grad_dmp_osc, res_grad_both
from ..res import res_grad_voigt, res_grad_thy

def res_scan_opt(p, *args) -> float:
    '''
    res_scan
    Scans minimal value of residual function with constraint
    ith value of parameter is fixed to p.

    Args:
     p: value of ith parameter
     args: arguments

           * 1st: i, index of parameter to scan
           * 2nd: parameter
           * 3rd: bounds
           * 4th: objective function which also gives its gradient
           * 5th to last: arguments for objective function
           * :math:`last-3`: fixed_param_idx
    
    Returns:
     residual value at params[i] = p
    '''
    param = np.atleast_1d(args[1]).copy()
    param[args[0]] = p
    bounds = len(args[2])*[None]
    for i in range(len(args[2])):
        bounds[i] = args[2][i]
    bounds[args[0]] = (p, p)
    func = args[3]
    fixed_param_idx = np.atleast_1d(args[-4]).copy()
    fixed_param_idx[args[0]] = True
    fargs_lst = (len(args)-4)*[None]
    for i in range(4, len(args)):
        fargs_lst[i-4] = args[i]
    fargs_lst[-4] = fixed_param_idx
    fargs = tuple(fargs_lst)
    res = minimize(func, param, args=fargs, bounds=bounds, method='L-BFGS-B', jac=True)

    return res['fun']

def ci_scan_opt_f(p, *args):
    '''
    Confidence interval scan with ith parameter is fixed to p. (for f-test based method)
    '''
    F_alpha, dfn, dfd, chi2_opt = args[:4]
    fargs = tuple(args[4:])
    return (res_scan_opt(p, *fargs)-chi2_opt/2)/dfn/(chi2_opt/(2*dfd))-F_alpha


class CIResult(dict):
    '''
    Class for represent confidence interval of each parameter

    Attributes:
     method ({'f'}): method to calculate confidance interval of each parameter
     alpha (float): significant level
     param_name (sequence of str): name of parameter
     x (np.ndarray): best parameter
     ci (sequence of tuple): confidence interval of each parameter at significant level alpha
    '''

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e
      
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
      
    def __dir__(self):
        return list(self.keys())
    
    def __str__(self):
        doc_lst = []
        doc_lst.append("[Report for Confidence Interval]")
        doc_lst.append(f"    Method: {self['method']}")
        doc_lst.append(f"    Significance level: {self['alpha']: 4e}")
        doc_lst.append(' ')
        doc_lst.append("[Confidence interval]")
        for pn, pv, ci in zip(self['param_name'], self['x'], self['ci']):
            if ci[0] != 0 and ci[1] != 0:
                tmp_doc_lst = []
                tmp_doc_lst.append(f"    {pv:.8f}".rstrip('0').rstrip('.'))
                tmp_doc_lst.append(f"- {-ci[0]: .8f}".rstrip('0').rstrip('.'))
                tmp_doc_lst.append(f"<= {pn} <=")
                tmp_doc_lst.append(f"{pv: .8f}".rstrip('0').rstrip('.'))
                tmp_doc_lst.append(f"+ {ci[1]: .8f}".rstrip('0').rstrip('.'))
                doc_lst.append(' '.join(tmp_doc_lst))
        return '\n'.join(doc_lst)



def is_better_fit(result1, result2) -> float:
    '''
    Compare fit based on f-test

    Args:
     result1 ({'StaticResult', 'TransientResult'}): fitting result class
      which has more parameter than result2
     result2 ({'StaticResult', 'TransientResult'}): fitting result class
      which has less parameter than result1

    Returns:
     p value of test, If p is smaller than your significant level,
     result1 is may better fit than result2.
     Otherwise, you cannot say resul1 is better fit than result2.
    
    Note:

     * The number of parameters in result1 should be greather than the number of parameters in result2.
     * The result1 and result2 should be different model for same data.

    '''
    chi2_1 = result1['chi2']; chi2_2 = result2['chi2']
    num_param_1 = result1['n_param']; num_param_2 = result2['n_param']
    num_pts_1 = result1['num_pts']; num_pts_2 = result2['num_pts']

    if num_param_1 <= num_param_2:
        raise Exception(f'Number of parameter in model 1: {num_param_1}' + 
        ' should be strictly greather than' + 
        f' the number of parameter in model 2: {num_param_2}')
    
    if num_pts_1 != num_pts_2:
        raise Exception('The result1 and result2 should be different model for same data')
    
    dfn = num_param_1 - num_param_2
    dfd = num_pts_1 - num_param_1

    F_test = (chi2_2-chi2_1)/dfn/(chi2_1/dfd)
    p = 1- f.cdf(F_test, dfn, dfd)
    return p

def confidence_interval(result, alpha: float) -> CIResult:
    '''
    Calculate 1d confidence interval of each parameter at significance level alpha
    Based on F-test method

    Args:
     result ({'StaticResult', 'TransientResult'}): fitting result class
     alpha (float): significance level

    Returns:
     CIResult class instance
    '''
    params = np.atleast_1d(result['x'])
    fix_param_idx = np.zeros(len(result['x']), dtype=bool)
    for i in range(params.size):
        fix_param_idx[i] = (result['bounds'][i][0] == result['bounds'][i][1])
    select_idx = fix_param_idx.copy()
    scan_idx = np.array(range(len(result['x'])))
    ci_lst = len(result['x'])*[(0, 0)]
    num_param = result['n_param']
    num_pts = result['num_pts']

    chi2_opt = result['chi2']
    dfn = 1; dfd = num_pts - num_param
    F_alpha = f.ppf(1-alpha, dfn, dfd)
    norm_alpha = np.ceil(norm.ppf(1-alpha/2))

    if result['model'] == 'decay':
        args = [F_alpha, dfn, dfd, chi2_opt, 0, params, result['bounds'], 
        res_grad_decay, result['n_decay'],
        result['base'], result['irf'], fix_param_idx, 
        result['t'], result['intensity'], result['eps']]
    elif result['model'] == 'dmp_osc':
        args = [F_alpha, dfn, dfd, chi2_opt, 0, params, result['bounds'],
        res_grad_dmp_osc, result['n_osc'], 
        result['irf'], fix_param_idx,
        result['t'], result['intensity'], result['eps']]
    elif result['model'] == 'both':
        args = [F_alpha, dfn, dfd, chi2_opt, 0, params, result['bounds'],
        res_grad_both,
        result['n_decay'], result['n_osc'], result['base'], result['irf'], fix_param_idx,
        result['t'], result['intensity'], result['eps']]
    elif result['model'] == 'voigt':
        args = [F_alpha, dfn, dfd, chi2_opt, 0, params, result['bounds'],
        res_grad_voigt,
        result['n_voigt'], result['edge'], result['n_edge'],
        result['base_order'], fix_param_idx,
        result['e'], result['intensity'], result['eps']]
    elif result['model'] == 'thy':
        args = [F_alpha, dfn, dfd, chi2_opt, 0, params, result['bounds'],
        res_grad_thy,
        result['policy'], result['thy_peak'], 
        result['edge'], result['n_edge'], 
        result['base_order'], fix_param_idx,
        result['e'], result['intensity'], result['eps']]
    
    sub_scan_idx = scan_idx[~select_idx]
    for idx in sub_scan_idx:
        p0 = params[idx]
        args[4] = idx
        fargs = tuple(args)
        p_eps = result['x_eps'][idx]
        p_lb = p0-norm_alpha*p_eps; p_ub = p0+norm_alpha*p_eps

        while ci_scan_opt_f(p_lb, *fargs) < 0:
            p_lb = p_lb - p_eps
            
        while ci_scan_opt_f(p_ub, *fargs) < 0:
            p_ub = p_ub + p_eps
            
        z1 = brenth(ci_scan_opt_f, p0, p_ub, args=fargs)
        z2 = brenth(ci_scan_opt_f, p_lb, p0, args=fargs)

        ci_lst[idx] = (z2-p0, z1-p0)
    
    ci_res = CIResult()
    ci_res['method'] = 'f'
    ci_res['alpha'] = alpha
    ci_res['param_name'] = result['param_name']
    ci_res['x'] = result['x']
    ci_res['ci'] = ci_lst
    return ci_res




    


'''
exp_conv_irf:
submodule for fitting data with sum of exponential decay or damped oscillation convolved with irf

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Tuple, Optional
import numpy as np
import scipy.linalg as LA

from .rate_eq import compute_signal_irf
from .A_matrix import make_A_matrix_exp, make_A_matrix_dmp_osc


def exp_conv(t: np.ndarray, fwhm: float, tau: np.ndarray,
             c: np.ndarray, base: Optional[bool] = True,
             irf: Optional[str] = 'g', eta: Optional[float] = None) -> np.ndarray:

    '''
    Constructs the model for the convolution of n exponential and
    instrumental response function
    Supported instrumental response function are

      * g: gaussian distribution
      * c: cauchy distribution
      * pv: pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       tau: life time for each component
       c: coefficient for each component
       base: whether or not include baseline [default: True]
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)

    Returns:
      Convolution of the sum of n exponential decays and instrumental
      response function.

    Note:
     Size of weight `c` is `num_comp+1` when base is set to true.
     Otherwise, its size is `num_comp`.
    '''
    A = make_A_matrix_exp(t, fwhm, tau, base, irf, eta)
    y = c@A

    return y


def fact_anal_exp_conv(t: np.ndarray,
                       fwhm: float,
                       tau: np.ndarray,
                       base: Optional[bool] = True,
                       irf: Optional[str] = 'g',
                       eta: Optional[float] = None,
                       intensity: Optional[np.ndarray] = None,
                       eps: Optional[np.ndarray] = None
                       ) -> np.ndarray:

    '''
    Estimate the best coefficiets when full width at half maximum fwhm
    and life constant tau are given
    
    When you fits your model to tscan data, you need to have
    good initial guess for not only life time of
    each component but also coefficients. To help this it solves
    linear least square problem to find best coefficients when fwhm and
    tau are given.

    Supported instrumental response functions are 

       1. 'g': gaussian distribution
       2. 'c': cauchy distribution
       3. 'pv': pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       tau: life time for each component
       base: whether or not include baseline [default: True]
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)
       intensity: intensity of time scan data to fit
       eps: standard error of data

    Returns:
     Best coefficient for given fwhm and tau, if base is set to `True` then
     size of coefficient is `num_comp + 1`, otherwise is `num_comp`.

    Note:
     the dimension of the intensity must be one.
    '''

    A = make_A_matrix_exp(t, fwhm, tau, base, irf, eta)
    if eps is None:
        eps = np.ones_like(intensity)
    
    y = intensity/eps
    A = np.einsum('j,ij->ij', 1/eps, A)
    c, _, _, _ = LA.lstsq(A.T, y, cond=1e-2)

    return c

def rate_eq_conv(t: np.ndarray, fwhm: float,
abs: np.ndarray, eigval: np.ndarray, V: np.ndarray, c: np.ndarray, 
irf: Optional[str] = 'g', eta: Optional[float] = None) -> np.ndarray:

    '''
    Constructs signal model rate equation with
    instrumental response function
    Supported instrumental response function are

      * g: gaussian distribution
      * c: cauchy distribution
      * pv: pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       abs: coefficient for each excited state
       eigval: eigenvalue of rate equation matrix 
       V: eigenvector of rate equation matrix 
       c: coefficient to match initial condition of rate equation
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)

    Returns:
      Convolution of the solution of the rate equation and instrumental
      response function.
    '''
    A = compute_signal_irf(t, eigval, V, c, fwhm, irf, eta)
    y = abs@A

    return y

def fact_anal_rate_eq_conv(t: np.ndarray, fwhm: float,
eigval: np.ndarray, V: np.ndarray, c: np.ndarray, 
exclude: Optional[str] = None, irf: Optional[str] = 'g',
eta: Optional[float] = None, intensity: Optional[np.ndarray] = None, 
eps: Optional[np.ndarray] = None) -> np.ndarray:

    '''
    Estimate the best coefficiets when full width at half maximum fwhm
    and eigenvector and eigenvalue of rate equation matrix are given

    Supported instrumental response functions are 

       1. 'g': gaussian distribution
       2. 'c': cauchy distribution
       3. 'pv': pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       eigval: eigenvalue of rate equation matrix 
       V: eigenvector of rate equation matrix 
       c: coefficient to match initial condition of rate equation
       exclude: exclude either 'first' or 'last' element or both 'first' and 'last' element.
                
                * 'first' : exclude first element
                * 'last' : exclude last element
                * 'first_and_last' : exclude both first and last element  
                * None : Do not exclude any element [default]
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)
       intensity: intensity of time scan data to fit
       eps: standard error of data

    Returns:
     Best coefficient for each component.

    Note:
     1. eigval, V, c should be obtained from solve_model
     2. The dimension of the intensity should be one.
    '''

    A = compute_signal_irf(t, eigval, V, c, fwhm, irf, eta)
    
    abs = np.zeros(A.shape[0])

    if eps is None:
        eps = np.ones_like(intensity)
    
    y = intensity/eps
    
    if exclude == 'first_and_last':
        B = np.einsum('j,ij->ij', 1/eps, A[1:-1,:])
    elif exclude == 'first':
        B = np.einsum('j,ij->ij', 1/eps, A[1:,:])
    elif exclude == 'last':
        B = np.einsum('j,ij->ij', 1/eps, A[:-1,:])
    else:
        B = np.einsum('j,ij->ij', 1/eps, A)
    
    coeff, _, _, _ = LA.lstsq(B.T, y, cond=1e-2)

    if exclude == 'first_and_last':
        abs[1:-1] = coeff
    elif exclude == 'first':
        abs[1:] = coeff
    elif exclude == 'last':
        abs[:-1] = coeff
    else:
        abs = coeff

    return abs


def dmp_osc_conv(t: np.ndarray, fwhm: float,
                      tau: np.ndarray,
                      T: np.ndarray,
                      phase: np.ndarray,
                      c: np.ndarray,
                      irf: Optional[str] = 'g',
                      eta: Optional[float] = None
                      ) -> np.ndarray:

    '''
    Constructs convolution of sum of damped oscillation and
    instrumental response function
    Supported instrumental response function are

      * g: gaussian distribution
      * c: cauchy distribution
      * pv: pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       tau: lifetime of vibration
       T: period of vibration
       phase: phase factor
       c: coefficient for each damping oscillation component
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)

    Returns:
      Convolution of sum of damped oscillation and instrumental
      response function.
    '''
    c_osc = np.zeros(2*c.size)
    for i in range(c.size):
      c_osc[i] = c[i]*np.cos(phase[i])
      c_osc[i+c.size] = -c[i]*np.sin(phase[i])

    A = make_A_matrix_dmp_osc(t, fwhm, tau, T, irf, eta)
    y = c_osc@A

    return y

def fact_anal_dmp_osc_conv(t: np.ndarray,
                       fwhm: float,
                       tau: np.ndarray, T: np.ndarray,
                       irf: Optional[str] = 'g',
                       eta: Optional[float] = None,
                       intensity: Optional[np.ndarray] = None,
                       eps: Optional[np.ndarray] = None
                       ) -> Tuple[np.ndarray, np.ndarray]:

    '''
    Estimate the best coefficiets and phase factor 
    when full width at half maximum fwhm
    , life constant tau and period of vibration T are given

    Supported instrumental response functions are 

       1. 'g': gaussian distribution
       2. 'c': cauchy distribution
       3. 'pv': pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       tau: life time for each component
       T: period of vibration of each component
       phase: phase factor for each component
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)
       intensity: intensity of time scan data to fit
       eps: standard error of data

    Returns:
     Pair of phase factor and Best coefficient for given damped oscillation component.

    Note:
     the dimension of the intensity should be one.
    '''
    
    A = make_A_matrix_dmp_osc(t, fwhm, tau, T, irf, eta)

    if eps is None:
        eps = np.ones_like(intensity)
    
    y = intensity/eps
    A = np.einsum('j,ij->ij', 1/eps, A)
    c, _, _, _ = LA.lstsq(A.T, y, cond=1e-2)
    c_cosine = c[:tau.size]
    c_sine = c[tau.size:]
    c_amp = np.sqrt(c_cosine**2+c_sine**2)
    phase_factor = -np.arctan2(c_sine, c_cosine)

    return phase_factor, c_amp

def sum_exp_dmp_osc_conv(t: np.ndarray, fwhm: float,
                         tau: np.ndarray,
                         tau_osc: np.ndarray,
                         T: np.ndarray,
                         phase: np.ndarray,
                         c: np.ndarray,
                         c_osc: np.ndarray,
                         base: Optional[bool] = True,
                         irf: Optional[str] = 'g',
                         eta: Optional[float] = None
                         ) -> np.ndarray:
    '''
    Constructs convolution of sum of exponential decay and damped oscillation and
    instrumental response function
    Supported instrumental response function are

      * g: gaussian distribution
      * c: cauchy distribution
      * pv: pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       tau: lifetime of decay
       tau_osc: lifetime of vibration
       T: period of vibration
       phase: phase factor
       c: coefficient for each decay component
       c_osc: coefficient for each vibrational component
       base: Whether or not use baseline feature for exponential decay component [default: True]
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)

    Returns:
      Convolution of sum of exponential decay and damped oscillation and instrumental
      response function.
    '''
    c_osc_2 = np.zeros(2*c_osc.size)
    for i in range(c_osc.size):
      c_osc_2[i] = c_osc[i]*np.cos(phase[i])
      c_osc_2[i+c_osc.size] = -c_osc[i]*np.sin(phase[i])

    A = make_A_matrix_exp(t, fwhm, tau, base, irf, eta)
    A_osc = make_A_matrix_dmp_osc(t, fwhm, tau_osc, T, irf, eta)
    y = c@A + c_osc_2@A_osc

    return y

def fact_anal_sum_exp_dmp_osc_conv(t: np.ndarray, fwhm: float,
                                   tau: np.ndarray, tau_osc: np.ndarray,
                                   T: np.ndarray,
                                   base: Optional[bool] = True,
                                   irf: Optional[str] = 'g',
                                   eta: Optional[float] = None,
                                   intensity: Optional[np.ndarray] = None,
                                   eps: Optional[np.ndarray] = None) -> \
                                    Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Estimate the best coefficiets and phase factor of oscillation part 
    when full width at half maximum fwhm
    , lifetime constant of decay component tau, lifetime constant of 
    oscillation component tau_osc and period of vibration T are given

    Supported instrumental response functions are 

       1. 'g': gaussian distribution
       2. 'c': cauchy distribution
       3. 'pv': pseudo voigt profile

    Args:
       t: time
       fwhm: full width at half maximum of instrumental response function
       tau: life time for each decay component
       tau_osc: life time for each vibration component
       base: Whether or not use baseline feature for exponential decay component
        [default: True]
       T: period of vibration of each component
       phase: phase factor for each component
       irf: shape of instrumental
            response function [default: g]

              * 'g': normalized gaussian distribution,
              * 'c': normalized cauchy distribution,
              * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`
       eta: mixing parameter for pseudo voigt profile
            (only needed for pseudo voigt profile)
       intensity: intensity of time scan data to fit
       eps: standard error of data

    Returns:
     Tuple of best coefficient and phase_factor for given decay and damped oscillation component.
     (c_(decay), phase_(osc), c_(osc))

    Note:
     the dimension of the intensity should be one.
    '''

    A = np.empty((tau.size+1*base+2*tau_osc.size, t.size))
    A[:tau.size+1*base, :] = make_A_matrix_exp(t, fwhm, tau, base, irf, eta)
    A[tau.size+1*base:, :] = make_A_matrix_dmp_osc(t, fwhm, tau_osc, T, irf, eta)

    if eps is None:
        eps = np.ones_like(intensity)
    
    y = intensity/eps
    A = np.einsum('j,ij->ij', 1/eps, A)
    c, _, _, _ = LA.lstsq(A.T, y, cond=1e-2)
    c_cosine = c[tau.size+1*base:tau.size+1*base+tau_osc.size]
    c_sine = c[tau.size+1*base+tau_osc.size:]
    c_osc = np.sqrt(c_cosine**2+c_sine**2)
    phase_osc = -np.arctan2(c_sine, c_cosine)

    return c[:tau.size+1*base], phase_osc, c_osc




# match_scale.py
# Wrapper script for match_scale()
# Date: 2022. 7. 25.
# Author: pistack
# Email: pistack@yonsei.ac.kr

import os
import sys

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack.tools._match_scale import match_scale

if __name__ == "__main__":
    match_scale()



import os
import sys
import unittest
import numpy as np

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import solve_seq_model, solve_l_model

class TestRateEqSolver(unittest.TestCase):

    def test_solve_seq_1(self):
        '''
        Test squential decay solver (A->B->C)
        '''
        tau1 = 1; tau2 = 10
        A = np.array([[-1/tau1, 0, 0],
        [1/tau1, -1/tau2, 0],
        [0, 1/tau2, 0]])
        eigval_ref = np.array([-1/tau1, -1/tau2, 0])
        y0 = np.array([1, 0, 0])
        eigval_tst, V_tst, c_tst = solve_seq_model(np.array([tau1, tau2]), y0)
        lV_tst = np.einsum('j,ij->ij', eigval_ref, V_tst)

        tst1 = np.allclose(eigval_ref, eigval_tst)
        tst2 = np.allclose(V_tst@c_tst, y0)
        tst3 = np.allclose(A@V_tst, lV_tst)
        self.assertEqual((tst1, tst2, tst3), (True, True, True))

    def test_solve_seq_2(self):
        '''
        Test squential decay solver (A->B->C->D)
        '''
        tau1 = 1; tau2 = 10; tau3 = 100
        A = np.array([[-1/tau1, 0, 0, 0],
        [1/tau1, -1/tau2, 0, 0],
        [0, 1/tau2, -1/tau3, 0],
        [0, 0, 1/tau3, 0]])
        eigval_ref = np.array([-1/tau1, -1/tau2, -1/tau3, 0])
        y0 = np.array([0.6, 0.1, 0.25, 0.05])
        eigval_tst, V_tst, c_tst = solve_seq_model(np.array([tau1, tau2, tau3]), y0)
        lV_tst = np.einsum('j,ij->ij', eigval_ref, V_tst)

        tst1 = np.allclose(eigval_ref, eigval_tst)
        tst2 = np.allclose(V_tst@c_tst, y0)
        tst3 = np.allclose(A@V_tst, lV_tst)
        self.assertEqual((tst1, tst2, tst3), (True, True, True))

    def test_solve_l_1(self):
        '''
        Test lower triangular rate equation solver (A->B, A->C)
        '''
        tau1 = 1; tau2 = 0.5
        A = np.array([[-(1/tau1+1/tau2), 0, 0],
        [1/tau1, 0, 0],
        [1/tau2, 0, 0]])
        eigval_ref = np.array([-(1/tau1+1/tau2), 0, 0])
        y0 = np.array([0.5, 0.3, 0.2])
        eigval_tst, V_tst, c_tst = solve_l_model(A, y0)
        lV_tst = np.einsum('j,ij->ij', eigval_ref, V_tst)

        tst1 = np.allclose(eigval_ref, eigval_tst)
        tst2 = np.allclose(V_tst@c_tst, y0)
        tst3 = np.allclose(A@V_tst, lV_tst)
        self.assertEqual((tst1, tst2, tst3), (True, True, True))

    def test_solve_l_2(self):
        '''
        Test lower triangular rate equation solver (A->B, C->D)
        '''
        tau1 = 1; tau2 = 0.5
        A = np.array([[-1/tau1, 0, 0, 0],
        [1/tau1, 0, 0, 0],
        [0, 0, -1/tau2, 0],
        [0, 0, 1/tau2, 0]])
        eigval_ref = np.array([-1/tau1, 0, -1/tau2, 0])
        y0 = np.array([0.5, 0, 0.5, 0])
        eigval_tst, V_tst, c_tst = solve_l_model(A, y0)
        lV_tst = np.einsum('j,ij->ij', eigval_ref, V_tst)

        tst1 = np.allclose(eigval_ref, eigval_tst)
        tst2 = np.allclose(V_tst@c_tst, y0)
        tst3 = np.allclose(A@V_tst, lV_tst)
        self.assertEqual((tst1, tst2, tst3), (True, True, True))

    def test_solve_l_3(self):
        '''
        Test lower triangular rate equation solver (A->B, A->C->)
        '''
        tau1 = 1; tau2 = 0.5; tau3 = 250
        A = np.array([[-(1/tau1+1/tau2), 0, 0],
        [1/tau1, 0, 0],
        [1/tau3, 0, -1/tau3]])
        eigval_ref = np.array([-(1/tau1+1/tau2), 0, -1/tau3])
        y0 = np.array([0.6, 0.3, 0.1])
        eigval_tst, V_tst, c_tst = solve_l_model(A, y0)
        lV_tst = np.einsum('j,ij->ij', eigval_ref, V_tst)

        tst1 = np.allclose(eigval_ref, eigval_tst)
        tst2 = np.allclose(V_tst@c_tst, y0)
        tst3 = np.allclose(A@V_tst, lV_tst)
        self.assertEqual((tst1, tst2, tst3), (True, True, True))
    
    def test_solve_l_4(self):
        '''
        Test lower triangular rate equation solver (A->B->, A->C->)
        '''
        tau1 = 1; tau2 = 0.5; tau3 = 250; tau4 = 500

        A = np.array([[-(1/tau1+1/tau2), 0, 0],
        [1/tau1, -1/tau3, 0],
        [1/tau2, 0, -1/tau4]])
        eigval_ref = np.array([-(1/tau1+1/tau2), -1/tau3, -1/tau4])
        y0 = np.array([0.6, 0.3, 0.1])
        eigval_tst, V_tst, c_tst = solve_l_model(A, y0)
        lV_tst = np.einsum('j,ij->ij', eigval_ref, V_tst)

        tst1 = np.allclose(eigval_ref, eigval_tst)
        tst2 = np.allclose(V_tst@c_tst, y0)
        tst3 = np.allclose(A@V_tst, lV_tst)
        self.assertEqual((tst1, tst2, tst3), (True, True, True))
    

if __name__ == '__main__':
    unittest.main()





import os
import sys
import unittest
import numpy as np
from scipy.signal import convolve

path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(path+"/../src/")

from TRXASprefitpack import calc_eta
from TRXASprefitpack import calc_fwhm
from TRXASprefitpack import gau_irf, cauchy_irf
from TRXASprefitpack import voigt
from TRXASprefitpack import dmp_osc_conv_pvoigt, dmp_osc_conv_gau, dmp_osc_conv_cauchy

def dmp_osc(t, k, T, phi):
    return np.exp(-k*t)*np.cos(2*np.pi/T*t+phi)*np.heaviside(t, 1)

class TestDmpOscConvIRF(unittest.TestCase):

    def test_dmp_osc_conv_gau(self):
        fwhm_G = 0.15; tau = 0.5; T = 0.3; phi = np.pi/4
        N_ref = 50000; N_sample = 100
        t = np.linspace(-2, 2, N_ref)
        t_sample = np.linspace(-1, 1, N_sample)
        sample_idx = np.searchsorted(t, t_sample)
        gau_ref = gau_irf(t, fwhm_G)
        dmp_osc_ref = dmp_osc(t, 1/tau, T, phi)
        ref = convolve(gau_ref, dmp_osc_ref, mode='same')*4/N_ref
        tst = dmp_osc_conv_gau(t_sample, fwhm_G, 1/tau, T, phi)
        cond = np.max(np.abs(ref[sample_idx]-tst))/np.max(ref) < 1e-3
        self.assertEqual(cond, True)

    def test_dmp_osc_conv_cauchy(self):
        fwhm_L = 0.15; tau = 0.5; T = 0.3; phi = np.pi/4
        N_ref = 50000; N_sample = 100
        t = np.linspace(-2, 2, N_ref)
        t_sample = np.linspace(-1, 1, N_sample)
        sample_idx = np.searchsorted(t, t_sample)
        cauchy_ref = cauchy_irf(t, fwhm_L)
        dmp_osc_ref = dmp_osc(t, 1/tau, T, phi)
        ref = convolve(cauchy_ref, dmp_osc_ref, mode='same')*4/N_ref
        tst = dmp_osc_conv_cauchy(t_sample, fwhm_L, 1/tau, T, phi)
        cond = np.max(np.abs(ref[sample_idx]-tst))/np.max(ref) < 1e-3
        self.assertEqual(cond, True)

    def test_dmp_osc_conv_pvoigt(self):
        '''
        Test pseudo voigt approximation.
        '''
        fwhm_L = 0.05; fwhm_G = 0.15; tau = 0.5; T = 0.3; phi = np.pi/4
        fwhm = calc_fwhm(fwhm_G, fwhm_L); eta = calc_eta(fwhm_G, fwhm_L)
        N_ref = 5000; N_sample = 100
        t = np.linspace(-2, 2, N_ref)
        t_sample = np.linspace(-1, 1, N_sample)
        sample_idx = np.searchsorted(t, t_sample)
        voigt_ref = voigt(t, fwhm_G, fwhm_L)
        dmp_osc_ref = dmp_osc(t, 1/tau, T, phi)
        ref = convolve(voigt_ref, dmp_osc_ref, mode='same')*4/N_ref
        tst = dmp_osc_conv_pvoigt(t_sample, fwhm, eta, 1/tau, T, phi)
        cond = np.max(np.abs(ref[sample_idx]-tst))/np.max(ref) < 2e-2
        self.assertEqual(cond, True)

if __name__ == '__main__':
    unittest.main()



'''
_static_voigt:
submodule for static spectrum with the
sum of voigt function, edge function and baseline function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Sequence, Tuple
import numpy as np
from numpy.polynomial.legendre import legval
from .static_result import StaticResult
from scipy.optimize import basinhopping
from scipy.optimize import least_squares
from ..mathfun.peak_shape import voigt, edge_gaussian, edge_lorenzian
from ..mathfun.A_matrix import fact_anal_A
from ..res.parm_bound import set_bound_e0, set_bound_t0
from ..res.res_voigt import residual_voigt, res_grad_voigt

def fit_static_voigt(e0_init: np.ndarray, fwhm_G_init: np.ndarray, fwhm_L_init: np.ndarray,
                     edge: Optional[str] = None, 
                     edge_pos_init: Optional[np.ndarray] = None,
                     edge_fwhm_init: Optional[np.ndarray] = None,
                     base_order: Optional[int] = None,
                     do_glb: Optional[bool] = False, 
                     method_lsq: Optional[str] = 'trf',
                     kwargs_glb: Optional[dict] = None, 
                     kwargs_lsq: Optional[dict] = None,
                     bound_e0: Optional[Sequence[Tuple[float, float]]] = None, 
                     bound_fwhm_G: Optional[Sequence[Tuple[float, float]]] = None, 
                     bound_fwhm_L: Optional[Sequence[Tuple[float, float]]] = None,
                     bound_edge_pos: Optional[Sequence[Tuple[float, float]]] = None,
                     bound_edge_fwhm: Optional[Sequence[Tuple[float, float]]] = None,
                     e: Optional[np.ndarray] = None, 
                     intensity: Optional[np.ndarray] = None,
                     eps: Optional[np.ndarray] = None) -> StaticResult:
                      
      '''
      driver routine for fitting static spectrum with sum of voigt component, edge and 
      polynomial baseline.

      It separates linear and non-linear part of the optimization problem to solve non linear least sequare
      optimization problem efficiently.

      Moreover this driver uses two step method to search best parameter, its covariance and
      estimated parameter error.
                    
      Step 1. (basinhopping)
      Use global optimization to find rough global minimum of our objective function.
      In this stage, it use analytic gradient for scalar residual function.

      Step 2. (method_lsq)
      Use least squares optimization algorithm to refine global minimum of objective function and approximate covariance matrix.
      Because of linear and non-linear seperation scheme, the analytic jacobian for vector residual function is hard to optain.
      Thus, in this stage, it uses numerical jacobian. 
      
      Args:
       e0_init (np.ndarray): initial peak position of each voigt component
       fwhm_G_init (np.ndarray): initial gaussian part of fwhm parameter of each voigt component
       fwhm_L_init (np.ndarray): initial lorenzian part of fwhm parameter of each voigt component
       edge ({'g', 'l'}): type of edge function. If edge is not set, edge feature is not included.
       edge_pos_init (np.ndarray): initial edge position
       edge_fwhm_init (np.ndarray): initial fwhm parameter of edge
       do_glb (bool): Whether or not use global optimization algorithm. If True then basinhopping algorithm is used.
       method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                             minimization (refinement of global optimization solution)
       kwargs_glb: keyward arguments for global optimization solver
       kwargs_lsq: keyward arguments for least square optimization solver
       bound_e0 (sequence of tuple): boundary for each voigt componet. If upper and lower bound are same, 
        driver assumes that the parameter is fixed during optimization. If `bound_e0` is `None`, 
        the upper and lower bound are given by `set_bound_e0`.
       bound_fwhm_G (sequence of tuple): boundary for fwhm_G parameter. 
        If `bound_fwhm_G` is `None`, the upper and lower bound are given as `(fwhm_G/2, 2*fwhm_G)`.
       bound_fwhm_L (sequence of tuple): boundary for fwhm_L parameter. 
        If `bound_fwhm_L` is `None`, the upper and lower bound are given as `(fwhm_L/2, 2*fwhm_L)`.
       bound_edge_pos (sequence of tuple): boundary for edge position, 
        if `bound_edge_pos` is `None` and `edge` is set, the upper and lower bound are given by `set_bound_t0`.
       bound_edge_fwhm (sequence of tuple): boundary for fwhm parameter of edge feature. 
        If `bound_edge_fwhm` is `None`, the upper and lower bound are given as `(edge_fwhm/2, 2*edge_fwhm)`.
       e (np.narray): energy range for data
       intensity (np.ndarray): intensity of static spectrum data
       eps (np.ndarray): estimated errors of static spectrum data

       Returns:
        StaticResult class object
       Note:
       
        * if initial fwhm_G is zero then such voigt component is treated as lorenzian component
        * if initial fwhm_L is zero then such voigt component is treated as gaussian component
      '''

      if e0_init is None:
            num_voigt = 0
            num_param = 0
      else:
            num_voigt = e0_init.size
            num_param = 3*num_voigt

      num_comp = num_voigt; num_edge = 0
      if edge is not None:
            num_edge = edge_pos_init.size
            num_comp = num_comp+num_edge
            num_param = num_param+2*num_edge
      
      if base_order is not None:
            num_comp = num_comp + base_order + 1
      
      param = np.empty(num_param, dtype=float)
      fix_param_idx = np.empty(num_param, dtype=bool)

      param[:num_voigt] = e0_init
      param[num_voigt:2*num_voigt] = fwhm_G_init
      param[2*num_voigt:3*num_voigt] = fwhm_L_init
      if edge is not None:
            param[3*num_voigt:3*num_voigt+num_edge] = edge_pos_init
            param[3*num_voigt+num_edge:] = edge_fwhm_init
      
      bound = num_param*[None]

      if bound_e0 is None:
            for i in range(num_voigt):
                  bound[i] = set_bound_e0(e0_init[i], fwhm_G_init[i], fwhm_L_init[i])
      else:
            bound[:num_voigt] = bound_e0
                      
      if bound_fwhm_G is None:
            for i in range(num_voigt):
                  bound[i+num_voigt] = (fwhm_G_init[i]/2, 2*fwhm_G_init[i])
      else:
            bound[num_voigt: 2*num_voigt] = bound_fwhm_G
      
      if bound_fwhm_L is None:
            for i in range(num_voigt):
                  bound[i+2*num_voigt] = (fwhm_L_init[i]/2, 2*fwhm_L_init[i])
      else:
            bound[2*num_voigt:3*num_voigt] = bound_fwhm_L
      
      if edge is not None:
            if bound_edge_pos is None:
                  for i in range(num_edge):
                        bound[3*num_voigt+i] = \
                              set_bound_t0(edge_pos_init[i], edge_fwhm_init[i])
            else:
                  bound[3*num_voigt:3*num_voigt+num_edge] = bound_edge_pos
            if bound_edge_fwhm is None:
                  for i in range(num_edge):
                        bound[3*num_voigt+num_edge+i] = \
                              (edge_fwhm_init[i]/2, 2*edge_fwhm_init[i])
            else:
                  bound[3*num_voigt+num_edge:] = bound_edge_fwhm

      for i in range(num_param):
            fix_param_idx[i] = (bound[i][0] == bound[i][1])
      
      if do_glb:
            go_args = (num_voigt, edge, num_edge, base_order, fix_param_idx, 
            e, intensity, eps)
            min_go_kwargs = {'args': go_args, 'jac': True, 'bounds': bound}
            if kwargs_glb is not None:
                  minimizer_kwargs = kwargs_glb.pop('minimizer_kwargs', None)
                  if minimizer_kwargs is None:
                        kwargs_glb['minimizer_kwargs'] = min_go_kwargs
                  else:
                        minimizer_kwargs['args'] = go_args
                        minimizer_kwargs['jac'] = True
                        minimizer_kwargs['bouns'] = bound
                        kwargs_glb['minimizer_kwargs'] = minimizer_kwargs
            else:
                  kwargs_glb = {'minimizer_kwargs' : min_go_kwargs}
            res_go = basinhopping(res_grad_voigt, param, **kwargs_glb)
      else:
            res_go = dict()
            res_go['x'] = param
            res_go['message'] = None
            res_go['nfev'] = 0

      param_gopt = res_go['x']

      lsq_args = (num_voigt, edge, num_edge, base_order, e, intensity, eps)

      if kwargs_lsq is not None:
            _ = kwargs_lsq.pop('args', None)
            _ = kwargs_lsq.pop('kwargs', None)
            kwargs_lsq['args'] = lsq_args
      else:
            kwargs_lsq = {'args' : lsq_args}

      bound_tuple = (num_param*[None], num_param*[None])
      for i in range(num_param):
            bound_tuple[0][i] = bound[i][0]
            bound_tuple[1][i] = bound[i][1]
            if bound[i][0] == bound[i][1]:
                  if bound[i][0] > 0:
                        bound_tuple[1][i] = bound[i][1]*(1+1e-8)+1e-16
                  else:
                        bound_tuple[1][i] = bound[i][1]*(1-1e-8)+1e-16
      
      #jacobian of vector residual function is inaccurate
      res_lsq = least_squares(residual_voigt, param_gopt, method=method_lsq, bounds=bound_tuple, **kwargs_lsq)
      param_opt = res_lsq['x']

      e0_opt = param_opt[:num_voigt]
      fwhm_G_opt = param_opt[num_voigt:2*num_voigt]
      fwhm_L_opt = param_opt[2*num_voigt:3*num_voigt]
      e0_edge_opt = param_opt[3*num_voigt:3*num_voigt+num_edge]
      fwhm_edge_opt = param_opt[3*num_voigt+num_edge:]


    # Calc individual chi2
      chi = res_lsq['fun']
      num_param_tot = num_comp+num_param-np.sum(fix_param_idx)
      chi2 = 2*res_lsq['cost']
      red_chi2 = chi2/(chi.size-num_param_tot)

      param_name = np.empty(param_opt.size, dtype=object)
      for i in range(num_voigt):
            param_name[i] = f'e0_{i+1}'
            param_name[num_voigt+i] = f'fwhm_(G, {i+1})'
            param_name[2*num_voigt+i] = f'fwhm_(L, {i+1})'
      
      if edge is not None:
            for i in range(num_edge):
                  param_name[3*num_voigt+i] = f'E0_({edge}, {i+1})'
            if edge == 'g':
                  for i in range(num_edge):
                        param_name[3*num_voigt+num_edge+i] = \
                              f'fwhm_(G, edge, {i+1})'
            elif edge == 'l':
                  for i in range(num_edge):
                        param_name[3*num_voigt+num_edge+i] = \
                              f'fwhm_(L, edge, {i+1})'
      
      A = np.empty((num_comp, e.size))
      
      for i in range(num_voigt):
            A[i, :] = voigt(e-e0_opt[i], fwhm_G_opt[i], fwhm_L_opt[i])
      
      base_start = num_voigt
      
      if edge is not None:
            base_start = base_start+num_edge
            if edge == 'g':
                  for i in range(num_edge):
                        A[num_voigt+i, :] = \
                              edge_gaussian(e-e0_edge_opt[i], fwhm_edge_opt[i])
            elif edge == 'l':
                  for i in range(num_edge):
                        A[num_voigt+i,:] = \
                              edge_lorenzian(e-e0_edge_opt[i], fwhm_edge_opt[i])
    
      if base_order is not None:
            e_max = np.max(e); e_min = np.min(e)
            e_norm = 2*(e-(e_max+e_min)/2)/(e_max-e_min)
            tmp = np.eye(base_order+1)
            A[base_start:, :] = legval(e_norm, tmp, tensor=True)
      
      c = fact_anal_A(A, intensity, eps)

      fit = c@A

      if edge is not None:
            fit_comp = np.einsum('i,ij->ij', c[:num_voigt+num_edge], 
            A[:num_voigt+num_edge,:])
      else:
            fit_comp = np.einsum('i,ij->ij', c[:num_voigt], 
            A[:num_voigt, :])
      
      base = None

      if base_order is not None:
            base = c[base_start:]@A[base_start:,:]
            
      res = intensity - fit
      
      jac = res_lsq['jac']
      hes = jac.T @ jac
      cov = np.zeros_like(hes)
      n_free_param = np.sum(~fix_param_idx)
      mask_2d = np.einsum('i,j->ij', ~fix_param_idx, ~fix_param_idx)
      cov[mask_2d] = np.linalg.inv(hes[mask_2d].reshape((n_free_param, n_free_param))).flatten()
      cov_scaled = red_chi2*cov
      param_eps = np.sqrt(np.diag(cov_scaled))
      corr = cov_scaled.copy()
      weight = np.einsum('i,j->ij', param_eps, param_eps)
      corr[mask_2d] = corr[mask_2d]/weight[mask_2d]

      result = StaticResult()
      result['model'] = 'voigt'
      result['e'] = e; result['intensity'] = intensity; result['eps'] = eps
      result['fit'] = fit; result['fit_comp'] = fit_comp 
      result['res'] = res; result['base_order'] = base_order
      result['edge'] = edge; result['n_voigt'] = num_voigt; result['n_edge'] = num_edge
      result['param_name'] = param_name; result['x'] = param_opt
      result['bounds'] = bound; result['base'] = base; result['c'] = c
      result['chi2'] = chi2
      result['aic'] = chi.size*np.log(chi2/chi.size)+2*num_param_tot
      result['bic'] = chi.size*np.log(chi2/chi.size)+num_param_tot*np.log(chi.size)
      result['red_chi2'] = red_chi2
      result['nfev'] = res_go['nfev'] + res_lsq['nfev']
      result['n_param'] = num_param_tot
      result['num_pts'] = chi.size; result['jac'] = jac
      result['cov'] = cov; result['corr'] = corr; result['cov_scaled'] = cov_scaled
      result['x_eps'] = param_eps
      result['method_lsq'] = method_lsq
      result['message_lsq'] = res_lsq['message']
      result['success_lsq'] = res_lsq['success']

      if result['success_lsq']:
            result['status'] = 0
      else:
            result['status'] = -1
      
      if do_glb:
            result['method_glb'] = 'basinhopping'
            result['message_glb'] = res_go['message'][0]  
      else:
            result['method_glb'] = None
            result['message_glb'] = None

      return result


'''
res_thy:
submodule for residual function and dfient for fitting static spectrum with the
sum of voigt broadened theoretical spectrum, edge function and base function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''
from typing import Optional, Sequence
import numpy as np
from numpy.polynomial.legendre import legval
from ..mathfun.A_matrix import fact_anal_A
from ..mathfun.peak_shape import voigt_thy, edge_gaussian, edge_lorenzian
from ..mathfun.peak_shape import deriv_voigt_thy, deriv_edge_gaussian, deriv_edge_lorenzian

def residual_thy(x0: np.ndarray, policy: str, thy_peak: Sequence[np.ndarray], 
                 edge: Optional[str] = None, num_edge: Optional[int] = 0,
                 base_order: Optional[int] = None, 
                 e: np.ndarray = None, 
                 intensity: np.ndarray = None, eps: np.ndarray = None) -> np.ndarray:
    '''
    residaul_thy
    `scipy.optimize.least_squares` compatible vector residual function for fitting static spectrum with the 
    sum of voigt broadend theoretical spectrum, edge function base function

    Args:
     x0: initial parameter

      * 1st and 2nd: fwhm_G and fwhm_L

      if policy == 'scale':

      * 3rd to :math:`2+{num}_{thy}` : peak_scale

      if policy == 'shift':

      * 3rd to :math:`2+{num}_{thy}` : peak_shift

      if policy == 'both':

      * 3rd to :math:`2+{num}_{thy}` : peak_shift
      * :math:`2+{num}_{thy}` to :math:`2+2 {num}_{thy}`: peak_scale

      if edge is not None:
      
      * :math:`2+{num}_{thy}+{num}_{edge}+i` or :math:`2+2{num_thy}+{num}_{edge}+i`: i th edge position
      * :math:`2+{num}_{thy}+2{num}_{edge}+i` or :math:`2+2{num_thy}+2{num}_{edge}+i`: fwhm of i th edge

     policy ({'shift', 'scale', 'both'}): Policy to match discrepency 
      between experimental data and theoretical spectrum.

      * 'shift' : Default option, shift peak position by peak_factor
      * 'scale' : scale peak position by peak_factor
      * 'both' : both shift and scale peak postition, peak_factor should be a tuple of `shift_factor` and `scale_factor`.

     thy_peak: theoretically calculated peak position and intensity
     edge ({'g', 'l'}): type of edge shape function
      if edge is not set, it does not include edge function.
     num_edge: the number of edge feature
     base_order (int): polynomial order of baseline function
      if base_order is not set, it does not include baseline function.
     fix_param_idx: idx for fixed parameter (masked array for `x0`)
     e: 1d array of energy points of data (n,)
     intensity: intensity of static data (n,)
     eps: estimated error of data (n,)

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    
    Note:
     * If fwhm_G of ith voigt component is zero then it is treated as lorenzian function with fwhm_L
     * If fwhm_L of ith voigt component is zero then it is treated as gaussian function with fwhm_G

    '''
    x0 = np.atleast_1d(x0)

    thy_comp = len(thy_peak)
    tot_comp = thy_comp

    if policy in ['scale', 'shift']:
        peak_factor = x0[2:2+thy_comp]
    elif policy == 'both':
        peak_factor = np.array(thy_comp, dtype=object)
        for i in range(thy_comp):
            peak_factor[i] = np.array([x0[2+i], x0[2+thy_comp+i]])
    
    if edge is not None:
        tot_comp = tot_comp+num_edge
    if base_order is not None:
        tot_comp = tot_comp+base_order+1
    
    A = np.empty((tot_comp, e.size))
    for i in range(thy_comp):
        A[i, :] = \
            voigt_thy(e, thy_peak[i], x0[0], x0[1], peak_factor[i], policy)/np.sum(thy_peak[i][:, 1])
    
    base_start = thy_comp
    if edge is not None:
        base_start = base_start+num_edge
        if policy in ['shift', 'scale']:
            param_edge_start = 2+thy_comp
        else:
            param_edge_start = 2+2*thy_comp
        if edge == 'g':
            for i in range(num_edge):
                A[thy_comp+i, :] = edge_gaussian(e-x0[param_edge_start+i], 
                x0[param_edge_start+num_edge+i])
        elif edge == 'l':
            for i in range(num_edge):
                A[thy_comp+i, :] = edge_lorenzian(e-x0[param_edge_start+i], 
                    x0[param_edge_start+num_edge+i])
    
    if base_order is not None:
        e_max = np.max(e); e_min = np.min(e)
        e_norm = 2*(e-(e_max+e_min)/2)/(e_max-e_min)
        tmp = np.eye(base_order+1)
        A[base_start:, :] = legval(e_norm, tmp, tensor=True)
    
    c = fact_anal_A(A, intensity, eps)
    chi = (c@A - intensity)/eps

    return chi

def res_grad_thy(x0: np.ndarray, policy: str, thy_peak: Sequence[np.ndarray], 
                 edge: Optional[str] = None, num_edge: Optional[int] = 0,
                 base_order: Optional[int] = None, 
                 fix_param_idx: Optional[np.ndarray] = None,
                 e: np.ndarray = None, 
                 intensity: np.ndarray = None, eps: np.ndarray = None) -> np.ndarray:
    '''
    res_grad_thy
    `scipy.optimize.minimize` compatible scalar residual function and its gradient for fitting static spectrum with the 
    sum of voigt broadend theoretical spectrum, edge function base function

    Args:
     x0: initial parameter

      * 1st and 2nd: fwhm_G and fwhm_L

      if policy == 'scale':

      * 3rd to :math:`2+{num}_{thy}` : peak_scale

      if policy == 'shift':

      * 3rd to :math:`2+{num}_{thy}` : peak_shift

      if policy == 'both':

      * 3rd to :math:`2+{num}_{thy}` : peak_shift
      * :math:`2+{num}_{thy}` to :math:`2+2 {num}_{thy}`: peak_scale

      if edge is not None:
      
      * :math:`2+{num}_{thy}+{num}_{edge}+i` or :math:`2+2{num_thy}+{num}_{edge}+i`: i th edge position
      * :math:`2+{num}_{thy}+2{num}_{edge}+i` or :math:`2+2{num_thy}+2{num}_{edge}+i`: fwhm of i th edge

     policy ({'shift', 'scale', 'both'}): Policy to match discrepency 
      between experimental data and theoretical spectrum.

      * 'shift' : Default option, shift peak position by peak_factor
      * 'scale' : scale peak position by peak_factor
      * 'both' : both shift and scale peak postition, peak_factor should be a tuple of `shift_factor` and `scale_factor`.

     thy_peak: theoretically calculated peak position and intensity
     edge ({'g', 'l'}): type of edge shape function
      if edge is not set, it does not include edge function.
     num_edge: the number of edge feature
     base_order (int): polynomial order of baseline function
      if base_order is not set, it does not include baseline function.
     fix_param_idx: idx for fixed parameter (masked array for `x0`)
     e: 1d array of energy points of data (n,)
     intensity: intensity of static data (n,)
     eps: estimated error of data (n,)

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    
    Note:
     * If fwhm_G of ith voigt component is zero then it is treated as lorenzian function with fwhm_L
     * If fwhm_L of ith voigt component is zero then it is treated as gaussian function with fwhm_G

    '''
    x0 = np.atleast_1d(x0)

    thy_comp = len(thy_peak)
    tot_comp = thy_comp
    area = np.empty(thy_comp)
    for i in range(thy_comp):
        area[i] = np.sum(thy_peak[i][:, 1])

    if policy in ['scale', 'shift']:
        peak_factor = x0[2:2+thy_comp]
    elif policy == 'both':
        peak_factor = np.array(thy_comp, dtype=object)
        for i in range(thy_comp):
            peak_factor[i] = np.array([x0[2+i], x0[2+thy_comp+i]])
    
    if edge is not None:
        tot_comp = tot_comp+num_edge
    if base_order is not None:
        tot_comp = tot_comp+base_order+1
    
    A = np.empty((tot_comp, e.size))
    for i in range(thy_comp):
        A[i, :] = \
            voigt_thy(e, thy_peak[i], x0[0], x0[1], peak_factor[i], policy)/area[i]
    
    base_start = thy_comp
    if edge is not None:
        base_start = base_start+num_edge
        if policy in ['shift', 'scale']:
            param_edge_start = 2+thy_comp
        else:
            param_edge_start = 2+2*thy_comp
        if edge == 'g':
            for i in range(num_edge):
                A[thy_comp+i, :] = edge_gaussian(e-x0[param_edge_start+i], 
                x0[param_edge_start+num_edge+i])
        elif edge == 'l':
            for i in range(num_edge):
                A[thy_comp+i, :] = edge_lorenzian(e-x0[param_edge_start+i], 
                    x0[param_edge_start+num_edge+i])
    
    if base_order is not None:
        e_max = np.max(e); e_min = np.min(e)
        e_norm = 2*(e-(e_max+e_min)/2)/(e_max-e_min)
        tmp = np.eye(base_order+1)
        A[base_start:, :] = legval(e_norm, tmp, tensor=True)
    
    c = fact_anal_A(A, intensity, eps)
    chi = (c@A - intensity)/eps
    df = np.zeros((intensity.size, x0.size))

    for i in range(thy_comp):
        deriv_thy = (c[i]/area[i])*\
            deriv_voigt_thy(e, thy_peak[i], x0[0], x0[1], peak_factor[i], policy)
        df[:, :2] = df[:, :2] + deriv_thy[:, :2]
        if policy in ['scale', 'shift']:
            df[:, 2+i] = deriv_thy[:, 2]
        else:
            df[:, 2+i] = deriv_thy[:, 2]
            df[:, 2+thy_comp+i] = deriv_thy[:, 3] 

    if edge is not None:
        if edge == 'g':
            for i in range(num_edge):
                df_edge = c[thy_comp+i]*deriv_edge_gaussian(e-x0[param_edge_start+i], 
                x0[param_edge_start+num_edge+i])
                df[:, param_edge_start+i] = -df_edge[:, 0]
                df[:, param_edge_start+num_edge+i] = df_edge[:, 1]
        elif edge == 'l':
            for i in range(num_edge):
                df_edge = c[thy_comp+i]*deriv_edge_lorenzian(e-x0[param_edge_start+i], 
                x0[param_edge_start+num_edge+i]) 
                df[:, param_edge_start+i] = -df_edge[:, 0]
                df[:, param_edge_start+num_edge+i] = df_edge[:, 1]
    
    df = np.einsum('i,ij->ij', 1/eps, df)

    df[:, fix_param_idx] = 0

    return np.sum(chi**2)/2, chi@df


# calc_sads
# Calculate species associated difference spectrum from experimental 
# energy scan data and the convolution of lower triangular 1st order rate equation model and
# instrumental response function.

import argparse
import numpy as np
import matplotlib.pyplot as plt

from ..mathfun.irf import calc_eta, calc_fwhm
from ..mathfun import solve_seq_model, solve_l_model
from ..driver import sads
from .misc import parse_matrix

description = '''
calc sads: Calculate species associated difference spectrum from experimental energy scan data and
the convolution of lower triangular 1st order rate equation model and instrumental response function

In rate equation model, the ground state would be
1. ''first_and_last'' species
2. ''first'' species
3. ''last'' species
4. ground state is not included in the rate equation model
'''

epilog = '''
*Note
1. Associated difference spectrum of ground state species is zero.
2. The rate equation matrix shoule be lower triangular.
3. Rate equation matrix for sequential decay model is sparse, so if your model is sequential decay then use --seq option
4. if you set shape of irf to pseudo voigt (pv), then you should provide two full width at half maximum value for gaussian and cauchy parts, respectively.
'''

gs_help = '''
Index of ground state species.
1. ``first_and_last``, first and last species are both ground state
2. ``first``, first species is ground state
3. ``last``,  last species is ground state
4. Did not set., There is no ground state species in model equation.
'''

seq_help = '''
Use sequential decay dynamics instead of more general lower triangular one.
If this option is turned on, it use following sequential decay dynamics
1 -> 2 -> 3 -> ... -> last
You can control the behavior of first and last species via --gsi option
'''

rate_eq_mat_help = '''
Filename for user supplied rate equation matrix. 
i th rate constant should be denoted by ki in rate equation matrix file.
Moreover rate equation matrix should be lower triangular.
'''

irf_help = '''
shape of instrument response functon
g: gaussian distribution
c: cauchy distribution
pv: pseudo voigt profile, linear combination of gaussian distribution and cauchy distribution 
    pv = eta*c+(1-eta)*g 
    the uniform fwhm parameter and 
    mixing parameter are determined according to Journal of Applied Crystallography. 33 (6): 1311–1316. 
'''

fwhm_G_help = '''
full width at half maximum for gaussian shape
It would not be used when you set cauchy irf function
'''

fwhm_L_help = '''
full width at half maximum for cauchy shape
It would not be used when you did not set irf or use gaussian irf function
'''



def calc_sads():

    tmp = argparse.RawTextHelpFormatter
    parser = argparse.ArgumentParser(formatter_class=tmp,
                                     description=description,
                                     epilog=epilog)
    parser.add_argument('-re_mat', '--rate_eq_mat', type=str, 
                        help=rate_eq_mat_help)
    parser.add_argument('--seq', action='store_true',
    help=seq_help)
    parser.add_argument('-gsi', '--gs_index', default=None, type=str, 
    choices=['first', 'last', 'first_and_last'], help=gs_help)
    parser.add_argument('--init_cond', default=None, nargs='+', type=float,
    help='initial condition ith argument is corresponding to inital concentration of ith component')
    parser.add_argument('--irf', default='g', choices=['g', 'c', 'pv'],
                        help=irf_help)
    parser.add_argument('--fwhm_G', type=float,
                        help=fwhm_G_help)
    parser.add_argument('--fwhm_L', type=float,
                        help=fwhm_L_help)
    parser.add_argument('escan_file',
                        help='filename for scale corrected energy scan file')
    parser.add_argument('escan_err_file',
    help='filename for the scaled estimated experimental error of energy scan file')
    parser.add_argument('-t0', '--time_zero', type=float,
                        help='time zero of energy scan')
    parser.add_argument('--escan_time', type=float, nargs='+',
    help='time delay for each energy scan')
    parser.add_argument('--tau', type=float, nargs='+',
                        help='lifetime of each decay path')
    parser.add_argument('-o', '--out', default='out',
                        help='prefix for output files')
    args = parser.parse_args()

    irf = args.irf
    if irf == 'g':
        if args.fwhm_G is None:
            raise Exception('You are using gaussian irf, so you should set fwhm_G!\n')
        else:
            fwhm = args.fwhm_G
            eta = 0
    elif irf == 'c':
        if args.fwhm_L is None:
            raise Exception('You are using cauchy/lorenzian irf,' +
            'so you should set fwhm_L!\n')
        else:
            fwhm = args.fwhm_L
            eta = 1
    else:
        if (args.fwhm_G is None) or (args.fwhm_L is None):
            raise Exception('You are using pseudo voigt irf,' +
            'so you should set both fwhm_G and fwhm_L!\n')
        else:
            fwhm = calc_fwhm(args.fwhm_G, args.fwhm_L)
            eta = calc_eta(args.fwhm_G, args.fwhm_L)

    if args.tau is None:
        raise Exception('Please set lifetime constants for each decay')
    else:
        tau = np.array(args.tau)
    
    if (args.time_zero is None):
        raise Exception('You should set time_zero for energy scan \n')
    else:
        time_zero = args.time_zero

    escan_data = np.genfromtxt(args.escan_file)
    escan_err = np.genfromtxt(args.escan_err_file)
    escan_time = np.array(args.escan_time)
    out_prefix = args.out
    exclude = args.gs_index
    y0 = np.array(args.init_cond)

    if args.seq:
        eigval, V, c = solve_seq_model(tau, y0)
    else:
        rate_eq_mat_str = np.genfromtxt(args.rate_eq_mat, dtype=str)
        L_mat = parse_matrix(rate_eq_mat_str, tau)
        eigval, V, c = solve_l_model(L_mat, y0)
    
    ads, ads_eps, fit = sads(escan_time-time_zero, fwhm, eigval, V, c, exclude, irf, eta,
    intensity=escan_data[:,1:], eps=escan_err)

    e = escan_data[:,0]

    out_ads = np.vstack((e, ads)).T
    out_fit = np.vstack((e,fit.T)).T
    ads_header_lst = []
    for i in range(ads.shape[0]):
        ads_header_lst.append(f'ex{i+1}')
    ads_header = '\t'.join(ads_header_lst)
    fit_header = '\t'.join(list(map(str, escan_time)))

    # save calculated sads results
    np.savetxt(f'{out_prefix}_sads.txt', out_ads, fmt=out_ads.shape[1]*['%.8e'], header='energy \t'+ads_header)
    np.savetxt(f'{out_prefix}_sads_eps.txt', ads_eps.T, fmt=ads_eps.shape[0]*['%.8e'], header=ads_header)
    np.savetxt(f'{out_prefix}_sads_fit.txt', out_fit, fmt=(len(escan_time)+1)*['%.8e'], header='energy \t'+fit_header)

    # plot sads results
    plt.figure(1)
    plt.title('Species Associated Difference Spectrum')
    for i in range(ads.shape[0]):
        plt.errorbar(e, ads[i,:], ads_eps[i,:], label=f'excited state {i+1}')
    plt.legend()

    offset = 2*np.max(np.abs(escan_data[:, 1:]))
    plt.figure(2)
    plt.title(f'Retrieved Energy Scan (time_zero: {time_zero:.3e}')
    for i in range(escan_time.size):
        plt.errorbar(e, escan_data[:, i+1]+i*offset, escan_err[:, i], marker='o', mfc='none',
        label=f'{escan_time[i]: .3e} (expt)')
        plt.plot(e, fit[:, i]+i*offset, label=f'{escan_time[i]: .3e} (fit)')
    plt.legend()
    plt.show()

    return